from routerl.keychain import Keychain as kc

from routerl.human_learning import (
    Gawron,
    Culo,
    WeightedAverage
)

def get_learning_model(params, initial_knowledge):
    """Returns a learning model based on the provided parameters.

    Args:
        params (dict): A dictionary containing model parameters.
        initial_knowledge (Any): A dictionary containing initial knowledge.
    Returns:
        BaseLearningModel: A learning model object.
    Raises:
        ValueError: If model is unknown.
    """

    model = params[kc.MODEL]
    if model == kc.GAWRON:
        return Gawron(params, initial_knowledge)
    elif model == kc.CULO:
        return Culo(params, initial_knowledge)
    elif model == kc.W_AVG:
        return WeightedAverage(params, initial_knowledge)
    else:
        raise ValueError('[MODEL INVALID] Unrecognized model: ' + model)


"""
This module contains the human and machine agent classes, which represent vehicles driving
from an origin to a destination in the simulation.
"""

import numpy as np

from abc import ABC, abstractmethod

from routerl.keychain import Keychain as kc
from routerl.human_learning import get_learning_model


class BaseAgent(ABC):
    """This is the abstract base class for the human and machine agent classes.

    Args:
        id (int): 
            The id of the agent.
        kind (str): 
            The kind of the agent (Human or AV).
        start_time (float): 
            The start time of the agent.
        origin (float): 
            The origin of the agent.
        destination (float): 
            The destination value of the agent.
        behavior (float): 
            The behavior of the agent.
    """

    def __init__(self, id, kind, start_time, origin, destination, behavior):
        self.id = id
        self.kind = kind
        self.start_time = start_time
        self.origin = origin
        self.destination = destination
        self.behavior = behavior
        self.last_action = 0

    @property
    @abstractmethod
    def last_reward(self) -> None:
        """Return the last reward of the agent.

        Returns:
            None
        """

        pass
    
    @last_reward.setter
    @abstractmethod
    def last_reward(self, reward) -> None:
        """Set the last reward of the agent.

        Args:
            reward (float): The reward of the agent.
        Returns:
            None
        """

        pass

    @abstractmethod
    def act(self, observation) -> None:
        """Pick action according to your knowledge, or randomly.

        Args:
            observation (float): The observation of the agent.
        Returns:
            None
        """

        pass

    @abstractmethod
    def learn(self, action, observation) -> None:
        """Pass the applied action and reward once the episode ends, and it will remember the consequences.

        Args:
            action (int): The action taken by the agent.
            observation (float): The observation of the agent.
        Returns:
            None
        """

        pass

    @abstractmethod
    def get_state(self, observation) -> None:
        """Return the state of the agent, given the observation

        Args:
            observation (float): The observation of the agent.
        Returns:
            None
        """

        pass

    @abstractmethod
    def get_reward(self, observation) -> None:
        """Derive the reward of the agent, given the observation

        Args:
            observation (float): The observation of the agent.
        Returns:
            None
        """

        pass


class HumanAgent(BaseAgent):
    """Class representing human drivers, responsible for modeling their learning process
    and decision-making in route selection.

    Args:
        id (int):
            The id of the agent.
        start_time (float):
            The start time of the agent.
        origin (float):
            The origin of the agent.
        destination (float):
            The destination value of the agent.
        params (dict):
            The parameters for the learning model of the agent as specified in `here <https://coexistence-project.github.io/RouteRL/documentation/pz_env.html#>`_.
        initial_knowledge (float):
            The initial knowledge of the agent.
    """

    def __init__(self, id, start_time, origin, destination, params, initial_knowledge):
        kind = kc.TYPE_HUMAN
        behavior = kc.SELFISH
        super().__init__(id, kind, start_time, origin, destination, behavior)
        self.model = get_learning_model(params, initial_knowledge)
        self.last_reward = None

    def __repr__(self):
        return f"Human {self.id}"

    @property
    def last_reward(self) -> float:
        """Set the last reward of the agent.

        Returns:
            float: The last reward of the agent.
        """

        return self._last_reward

    @last_reward.setter
    def last_reward(self, reward) -> None:
        """Set the last reward of the agent.

        Args:
            reward (float): The reward of the agent.
        Returns:
            None
        """

        self._last_reward = reward

    def act(self, observation) -> int:  
        """Returns the agent's action (route of choice) based on the current observation from the environment.

        Args:
            observation (list): The observation of the agent.
        Returns:
            int: The action of the agent.
        """

        return self.model.act(observation)

    def learn(self, action, observation) -> None:
        """Updates the agent's knowledge based on the action taken and the resulting observations.

        Args:
            action (int): The action of the agent.
            observation (list[dict]): The observation of the agent.
        Returns:
            None
        """

        reward = self.get_reward(observation)
        self.last_reward = reward
        self.model.learn(None, action, reward)

    def get_state(self, _) -> None:
        """Returns the current state of the agent.

        Args:
            _ (Any): The current state of the agent.
        Returns:
            None
        """

        return None

    def get_reward(self, observation: list[dict]) -> float:
        """This function calculated the reward of each individual agent.

        Args:
            observation (list[dict]): The observation of the agent.
        Returns:
            float: Own travel time of the agent.
        """

        own_tt = -1 * next(obs[kc.TRAVEL_TIME] for obs in observation if obs[kc.AGENT_ID] == self.id)
        return own_tt
    

class MachineAgent(BaseAgent):
    """A class that models Autonomous Vehicles (AVs), focusing on their learning mechanisms
    and decision-making processes for selecting optimal routes.

    Args:
        id (int): 
            The id of the agent.
        start_time (float): 
            The start time of the agent.
        origin (float): 
            The origin of the agent.
        destination (float): 
            The destination value of the agent.
        params (dict): 
            The parameters of the machine agent as specified in `here <https://coexistence-project.github.io/RouteRL/documentation/pz_env.html#>`_.
        action_space_size (int): 
            The size of the action space of the agent.
    """

    def __init__(self, id, start_time, origin, destination, params, action_space_size):
        kind = kc.TYPE_MACHINE
        behavior = params[kc.BEHAVIOR]
        super().__init__(id, kind, start_time, origin, destination, behavior)
        self.observed_span = params[kc.OBSERVED_SPAN]
        self.action_space_size = action_space_size
        self.state_size = action_space_size * 2
        self.model = None
        self.last_reward = None
        self.rewards_coefs = self._get_reward_coefs()

    def __repr__(self) -> str:
        machine_id = f"Machine {self.id}"

        return machine_id

    @property
    def last_reward(self) -> float:
        """Set the last reward of the agent.

        Returns:
            float: The last reward of the agent.
        """

        return self._last_reward

    @last_reward.setter
    def last_reward(self, reward) -> None:
        """Sets the last reward of the agent.

        Args:
            reward (float): The reward of the agent.
        Returns:
            None
        """

        self._last_reward = reward

    def act(self, _) -> None:
        """**Deprecated**

        Args:
            _ (Any): The current state of the agent.
        Returns:
            None
        """

        return None

    def learn(self, _) -> None:
        """**Deprecated**
        
        Args:
            _ (Any): The current state of the agent.
            
        Returns:
            None
        """

        return None

    def get_state(self, observation: list[dict]) -> list[int]:
        """Generates the current state representation based on recent observations of agents navigating
        from the same origin to the same destination.

        Args:
            observation (list[dict]): The recent observations of the agent.
            
        Returns:
            list[int]: The current state representation.
        """

        min_start_time = self.start_time - self.observed_span
        human_prior, machine_prior = list(), list()
        for obs in observation:
            if ((obs[kc.AGENT_ORIGIN], obs[kc.AGENT_DESTINATION]) == (self.origin, self.destination)
                    and (obs[kc.AGENT_START_TIME] > min_start_time)
            ):
                if obs[kc.AGENT_KIND] == kc.TYPE_HUMAN:
                    human_prior.append(obs)
                elif obs[kc.AGENT_KIND] == kc.TYPE_MACHINE:
                    machine_prior.append(obs)

        warmth_human = [0] * (self.state_size // 2)
        warmth_machine = [0] * (self.state_size // 2)
        
        if human_prior:
            for row in human_prior:
                action = row[kc.ACTION]
                start_time = row[kc.AGENT_START_TIME]
                warmth = start_time - min_start_time
                warmth_human[action] += warmth
                
        if machine_prior:
            for row in machine_prior:
                action = row[kc.ACTION]
                start_time = row[kc.AGENT_START_TIME]
                warmth = start_time - min_start_time
                warmth_machine[action] += warmth

        warmth_agents = warmth_human + warmth_machine
        return warmth_agents

    def get_reward(self, observation: list[dict]) -> float:
        """This method calculated the reward of each individual agent, based on the travel time of the agent,
        the group of agents, the other agents, and all agents, weighted according to the agent's behavior.

        Args:
            observation (list[dict]): The current observation of the agent.
                
        Returns:
            float: The reward of the agent.
        """

        min_start_time, max_start_time = self.start_time - self.observed_span, self.start_time + self.observed_span
        
        vicinity_obs = list()
        for obs in observation:
            if (obs[kc.AGENT_ORIGIN], obs[kc.AGENT_DESTINATION]) == (self.origin, self.destination):
                if min_start_time <= obs[kc.AGENT_START_TIME] <= max_start_time:
                    vicinity_obs.append(obs)

        group_obs, others_obs, all_obs, own_tt = list(), list(), list(), None
        for obs in vicinity_obs:
            all_obs.append(obs[kc.TRAVEL_TIME])
            if obs[kc.AGENT_KIND] == self.kind:
                group_obs.append(obs[kc.TRAVEL_TIME])
            else:
                others_obs.append(obs[kc.TRAVEL_TIME])
            if obs[kc.AGENT_ID] == self.id:
                own_tt = obs[kc.TRAVEL_TIME]
                
        group_tt = np.mean(group_obs) if group_obs else 0
        others_tt = np.mean(others_obs) if others_obs else 0
        all_tt = np.mean(all_obs) if all_obs else 0
        
        a, b, c, d = self.rewards_coefs
        agent_reward  = a * own_tt + b * group_tt + c * others_tt + d * all_tt
        return agent_reward

    def _get_reward_coefs(self) -> tuple:
        a, b, c, d = 0, 0, 0, 0
        if self.behavior == kc.SELFISH:
            a, b, c, d = -1, 0, 0, 0 
        elif self.behavior == kc.COMPETITIVE:
            a, b, c, d = -2, 0, 1, 0
        elif self.behavior == kc.COLLABORATIVE:
            a, b, c, d = -0.5, -0.5, 0, 0
        elif self.behavior == kc.SOCIAL:
            a, b, c, d = -0.5, 0, 0, -0.5
        elif self.behavior == kc.ALTRUISTIC:
            a, b, c, d = 0, 0, 0, -1
        elif self.behavior == kc.MALICIOUS:
            a, b, c, d = 0, 0, 1, 0
        return a, b, c, d


import logging
import os
import polars as pl

from routerl.keychain import Keychain as kc
from routerl.utilities import make_dir

logger = logging.getLogger()
logger.setLevel(logging.WARNING)


class Recorder:
    """Record the training process.

    Args:
        params (list):
            Plotter parameters as specified `here <https://coexistence-project.github.io/RouteRL/documentation/pz_env.html#>`_.
        
    Methods:
        record: saves episode data and detector statistics to disk.
        remember_episode: records episode data.
        remember_detector: decords detector data
    """

    def __init__(self, params):
        self.params = params
        self.records_folder = self.params[kc.RECORDS_FOLDER]

        self.episodes_folder = make_dir([self.records_folder, kc.EPISODES_LOGS_FOLDER])
        self.detector_folder = make_dir([self.records_folder, kc.DETECTOR_LOGS_FOLDER])

        self._clear_records(self.episodes_folder)
        self._clear_records(self.detector_folder)
        
        self.loss_file_path = self._get_txt_file_path(kc.LOSSES_LOG_FILE_NAME)
        logging.info(f"[SUCCESS] Recorder is now here to record!")

    ################################
    ######## Initial helpers #######
    ################################

    def _clear_records(self, folder) -> None:
        """Clears records from records_folder.

        Args:
            folder: folder to clear records from.
        Returns:
            None
        """

        if os.path.exists(folder):
            for file in os.listdir(folder):
                os.remove(os.path.join(folder, file))

    def _get_txt_file_path(self, filename) -> str:
        """Gets text file from records_folder.

        Args:
            filename: filename.
        Returns:
            None
        """

        log_file_path = make_dir(self.records_folder, filename)
        if os.path.exists(log_file_path):
            os.remove(log_file_path)
        return log_file_path

    ################################
    ####### Remember methods #######
    ################################

    def record(self, episode, ep_observations, cost_tables, det_dict) -> None:
        """Records the episode.

        Args:
            episode: episode.
            ep_observations: episode observations.
            rewards: rewards.
            cost_tables: cost_tables.
            det_dict: det_dict.
        Returns:
            None
        """

        self.remember_episode(episode, ep_observations, cost_tables)
        self.remember_detector(episode, det_dict)
        

    def remember_episode(self, episode, ep_observations, cost_tables) -> None:
        """Remember the episode.

        Args:
            episode: episode.
            ep_observations: episode observations.
            cost_tables: cost_tables.
        Returns:
            None
        """
        
        ep_observations_df = pl.from_dicts(ep_observations)

        for entry in cost_tables:
            entry['cost_table'] = ','.join(map(str, entry['cost_table']))

        cost_tables_df = pl.from_dicts(cost_tables)
        
        merged_df = ep_observations_df.join(cost_tables_df, on=kc.AGENT_ID)
        merged_df.write_csv(make_dir(self.episodes_folder, f"ep{episode}.csv"))

    def remember_detector(self, episode, det_dict) -> None:
        """Remember the detector.

        Args:
            episode: episode.
            det_dict: det_dict.
        Returns:
            None
        """
        
        df = pl.DataFrame(list(det_dict.items()), schema=['detid', 'flow'], orient="row")
        df.write_csv(make_dir(self.detector_folder, f'detector_ep{episode}.csv'))

    def save_losses(self, agents) -> None:
        """Save losses.

        Args:
            agents: agents.
        Returns:
            None
        """

        losses = list()
        for a in agents:
            loss = getattr(a.model, 'loss', None)
            if loss is not None:
                losses.append(loss)
        if len(losses):
            mean_losses = [0] * len(losses[-1])
            for loss in losses:
                for i, l in enumerate(loss):
                    mean_losses[i] += l
            mean_losses = [m / len(losses) for m in mean_losses]
            with open(self.loss_file_path, "w") as file:
                for m_l in mean_losses:
                    file.write(f"{m_l}\n")


import os

import logging
import numpy as np
import pandas as pd
import random

from routerl.keychain import Keychain as kc
from routerl.environment import HumanAgent
from routerl.environment import MachineAgent


logger = logging.getLogger()
logger.setLevel(logging.WARNING)


def generate_agents(params, free_flow_times, generate_data, seed=23423) -> list:
    """Generates agent objects for the enviornment.
    
    This function creates agents based on predefined parameters, either by 
    generating new data or loading existing data from a CSV file. Each agent 
    is instantiated as either a `MachineAgent` or `HumanAgent`, depending on 
    its type.

    Args:
        params (dict):
            Agent parameters dictionary as specified in `here <https://coexistence-project.github.io/RouteRL/documentation/pz_env.html#>`_.
        free_flow_times (dict[tuple[int], list[float]]):
            Free flow times of route options per OD.
        generate_data (bool):
            If True, generates new agent data. If False, loads existing agent 
            data from a CSV file specified by `params[kc.RECORDS_FOLDER]`.
        seed (int, optional): Random seed for reproducibility. Defaults to 23423.
        
    Returns:
        list: A list of `BaseAgent` objects (either `MachineAgent` or `HumanAgent`).
        
    Raises:
        ValueError: If an unrecognized agent type is encountered in the data.
    """

    set_seed(seed)
    if generate_data:
        # Generate agent data
        agents_data_df = generate_agent_data(params, seed)
    else:
        # Load agent data
        agents_csv_path = os.path.join(params[kc.RECORDS_FOLDER], kc.AGENTS_CSV_FILE_NAME)
        agents_data_df = pd.read_csv(agents_csv_path)

    # Getting parameters
    action_space_size = params[kc.ACTION_SPACE_SIZE]
    agents = list()     # Where we will store & return agents

    # Generating agent objects from generated agent data
    for _, row in agents_data_df.iterrows():
        row_dict = row.to_dict()

        id, start_time = row_dict[kc.AGENT_ID], row_dict[kc.AGENT_START_TIME]
        origin, destination = row_dict[kc.AGENT_ORIGIN], row_dict[kc.AGENT_DESTINATION]

        if row_dict[kc.AGENT_KIND] == kc.TYPE_MACHINE:
            agent_params = params[kc.MACHINE_PARAMETERS]
            mutate_to = MachineAgent(id,
                                     start_time,
                                     origin,
                                     destination,
                                     agent_params,
                                     action_space_size)
            agents.append(mutate_to)
        elif row_dict[kc.AGENT_KIND] == kc.TYPE_HUMAN:
            agent_params = params[kc.HUMAN_PARAMETERS]
            initial_knowledge = free_flow_times[(origin, destination)]
            agents.append(HumanAgent(id,
                                     start_time,
                                     origin,
                                     destination,
                                     agent_params,
                                     initial_knowledge))
        else:
            raise ValueError('[AGENT TYPE INVALID] Unrecognized agent type: ' + row_dict[kc.AGENT_KIND])
    return agents

def generate_agent_data(params, seed=23423) -> pd.DataFrame:
    """Generates agent data.

    Constructs a dataframe, where each row is an agent and columns are attributes.

    Args:
        params (dict): Agent parameters dictionary as specified in `here <https://coexistence-project.github.io/RouteRL/documentation/pz_env.html#>`_.
        seed (int): Random seed.
    Returns:
        agents_df (pd.DataFrame): Pandas dataframe with agents attributes.
    """
    
    num_agents = params[kc.NUM_AGENTS]
    simulation_timesteps = params[kc.SIMULATION_TIMESTEPS]
    agent_attributes = [kc.AGENT_ID, kc.AGENT_ORIGIN, kc.AGENT_DESTINATION, kc.AGENT_START_TIME, kc.AGENT_KIND]
    
    num_origins = len(params[kc.ORIGINS])
    num_destinations = len(params[kc.DESTINATIONS])
    
    rng = set_seed(seed)
    agents_df = pd.DataFrame(columns=agent_attributes)  # Where we store our agents
    
    mean_timestep = simulation_timesteps / 2
    std_dev_timestep = simulation_timesteps / 6

    # Generating agent data
    for id in range(num_agents):
        
        # Decide on agent type (temporary solution)
        agent_type = kc.TYPE_HUMAN

        # Randomly assign origin & destination
        origin, destination = random.randrange(num_origins), random.randrange(num_destinations)

        # Randomly assign start time (normal dist)
        start_time = int(rng.normal(mean_timestep, std_dev_timestep))
        start_time = max(0, min(simulation_timesteps, start_time))

        # Registering to the dataframe
        agent_features = [id, origin, destination, start_time, agent_type]
        agent_dict = {attribute : feature for attribute, feature in zip(agent_attributes, agent_features)}
        agents_df.loc[id] = agent_dict
        
    # Saving the generated agents to a csv file
    agents_csv_path = os.path.join(params[kc.RECORDS_FOLDER], kc.AGENTS_CSV_FILE_NAME)
    agents_csv_path = str(agents_csv_path)
    agents_df.to_csv(agents_csv_path, index=False)
    
    return agents_df

def set_seed(seed):
    """Set the seed for random number generation.

    Arguments:
        seed (int): Random seed
    Returns:
        rng: Random default rng for random number generation
    """

    random.seed(seed)
    np.random.seed(seed)
    rng = np.random.default_rng(seed)

    return rng



"""
This file, simulator.py, defines the SumoSimulator class, 
which manages the interaction between driver agents and the SUMO traffic simulator 
by handling simulation setup, path generation, vehicle management, and simulation control.
"""
import os
import janux as jx
import logging
import random
import pandas as pd
import traci

from routerl.keychain import Keychain as kc
from routerl.utilities import confirm_env_variable

logger = logging.getLogger()
logger.setLevel(logging.WARNING)


class SumoSimulator():
    """Sumo simulator class

    A class responsible for managing the communication between our learning agents and the SUMO traffic simulator.
    SUMO provides the traffic environment where vehicles travel between designated origins and destinations,
    and it returns the corresponding travel times for these vehicles.

    Args:
        params (dict):
            Dictionary of parameters for the SUMO environment. Specified `here <https://coexistence-project.github.io/RouteRL/documentation/pz_env.html#>`_.
        path_gen_params (dict):
            Dictionary of parameters for the SUMO environment. specified `here <https://coexistence-project.github.io/RouteRL/documentation/pz_env.html#>`_.
        seed (int): 
            Random seed for reproducibility.
        
    Attributes:
        network_name: Network name.
        simulation_length: Simulation length.
        sumo_id: SUMO connection id.
        sumo_connection: Traci-SUMO connection object.
        timestep: Time step being simulated within the day.
    """

    def __init__(self, params: dict, path_gen_params: dict, seed: int = 23423):
        self.network_name        = params[kc.NETWORK_NAME]
        self.sumo_type           = params[kc.SUMO_TYPE]
        self.number_of_paths     = params[kc.NUMBER_OF_PATHS]
        self.simulation_length   = params[kc.SIMULATION_TIMESTEPS]

        curr_dir = os.path.dirname(os.path.abspath(__file__))

        self.network_folder      = os.path.join(curr_dir,
                                                kc.NETWORK_FOLDER).replace("$net$", self.network_name)
        self.sumo_config_path    = os.path.join(curr_dir,
                                                kc.SUMO_CONFIG_PATH).replace("$net$", self.network_name)
        self.routes_xml_path     = os.path.join(curr_dir,
                                                kc.ROU_FILE_PATH).replace("$net$", self.network_name)
        self.sumo_fcd            = os.path.join(curr_dir,
                                                kc.SUMO_FCD).replace("$net$", self.network_name)
        self.detector_save_path  = os.path.join(curr_dir,
                                                kc.DETECTORS_CSV_PATH).replace("$net$", self.network_name)
        self.conn_file_path      = os.path.join(curr_dir,
                                                kc.CONNECTION_FILE_PATH).replace("$net$", self.network_name)
        self.edge_file_path      = os.path.join(curr_dir,
                                                kc.EDGE_FILE_PATH).replace("$net$", self.network_name)
        self.nod_file_path       = os.path.join(curr_dir,
                                                kc.NOD_FILE_PATH).replace("$net$", self.network_name)
        self.rou_xml_save_path   = os.path.join(curr_dir,
                                                kc.ROUTE_XML_PATH).replace("$net$", self.network_name)
        self.det_xml_save_path   = os.path.join(curr_dir,
                                                kc.DETECTORS_XML_PATH).replace("$net$", self.network_name)
        self.default_od_path     = os.path.join(curr_dir,
                                                kc.DEFAULT_ODS_PATH)
        
        self.paths_csv_file_path = os.path.join(params[kc.RECORDS_FOLDER], kc.PATHS_CSV_FILE_NAME)

        random.seed(seed)

        self.seed = seed
        self.sumo_id = f"{random.randint(0, 1000)}"
        self.sumo_connection = None

        confirm_env_variable(kc.ENV_VAR, append="tools")

        if path_gen_params is not None:
            self._get_paths(params, path_gen_params)
            logging.info("[SUCCESS] Path generation completed.")
        self._check_paths_ready()
        self.detectors_name = self._get_detectors()
        self.timestep = 0
        self.route_id_cache = dict()

        logging.info("[SUCCESS] Simulator is ready to simulate!")

    ################################
    ######## CONFIG CHECKS #########
    ################################

    def _check_paths_ready(self) -> None:

        if os.path.isfile(self.rou_xml_save_path):
            logging.info("[CONFIRMED] Paths file is ready.")
        else:
            raise FileNotFoundError(
                "Paths file is not ready. Please generate paths first.\n"
                "To do this, define arguments in params.json and pass it "
                "to the environment under path_generation_parameters"
            )
            
    def _get_paths(self, params: dict, path_gen_params: dict) -> None:

        # Build the network
        network = jx.build_digraph(self.conn_file_path, self.edge_file_path, self.routes_xml_path)
        
        # Get origins and destinations
        origins = path_gen_params[kc.ORIGINS]
        destinations = path_gen_params[kc.DESTINATIONS]
        
        # Generate paths
        path_gen_kwargs = {
            "number_of_paths": path_gen_params[kc.NUMBER_OF_PATHS],
            "random_seed": self.seed,
            "num_samples": path_gen_params[kc.NUM_SAMPLES],
            "beta": path_gen_params[kc.BETA],
            "weight": path_gen_params[kc.WEIGHT],
            "verbose": False
        }
        routes = jx.basic_generator(network, origins, destinations, as_df=True, calc_free_flow=True, **path_gen_kwargs)
        self._save_paths_to_disc(routes, origins, destinations)
        
        # Save paths visualizations
        path_visuals_path = params[kc.PLOTS_FOLDER]
        os.makedirs(path_visuals_path, exist_ok=True)
        # Visualize paths and save figures
        for origin_idx, origin in enumerate(origins):
            for dest_idx, destination in enumerate(destinations):
                # Filter routes for the current origin-destination pair
                routes_to_show = (routes[(routes["origins"] == origin_idx)
                                         & (routes["destinations"] == dest_idx)]['path'])
                routes_to_show = [route.split(" ") for route in routes_to_show]
                # Specify the save path and title for the figure
                fig_save_path = os.path.join(path_visuals_path, f"{origin_idx}_{dest_idx}.png")
                title=f"Origin: {origin_idx} ({origin}), Destination: {dest_idx} ({destination})"
                # Show the routes
                jx.show_multi_routes(self.nod_file_path, self.edge_file_path,
                                    routes_to_show, origin, destination, 
                                    show=False, save_file_path=fig_save_path, title=title)

    def _save_paths_to_disc(self, routes_df: pd.DataFrame, origins: list, destinations: list) -> None:

        origins = {node_name: idx for idx, node_name in enumerate(origins)}
        destinations = {node_name: idx for idx, node_name in enumerate(destinations)}
        
        # Format routes dataframe
        routes_df["origins"] = routes_df["origins"].apply(lambda x: origins[x])
        routes_df["destinations"] = routes_df["destinations"].apply(lambda x: destinations[x])
        routes_df["path"] = routes_df["path"].apply(lambda x: x.replace(",", " "))
        routes_df.sort_values(by=["origins", "destinations"], inplace=True)

        # Save paths to csv
        routes_df.to_csv(self.paths_csv_file_path, index=False)

        # Convert routes dataframe to a dictionary with od indices
        paths_dict = dict()
        for origin_idx in origins.values():
            for destination_idx in destinations.values():
                rows = routes_df[(routes_df["origins"] == origin_idx) & (routes_df["destinations"] == destination_idx)]
                paths = rows["path"].to_list()
                paths_dict[(origin_idx, destination_idx)] = paths

        # Save paths to xml
        with open(self.rou_xml_save_path, "w") as rou:
            print("""<routes>""", file=rou)
            # TODO: Following two lines are hardcoded. Change them to be dynamic.
            print("<vType id=\"Human\" color=\"red\" guiShape=\"passenger/sedan\"/>", file=rou)
            print("<vType id=\"AV\" color=\"yellow\"/>", file=rou)
            for origin_idx in origins.values():
                for destination_idx in destinations.values():
                    paths = (routes_df[(routes_df["origins"] == origin_idx)
                                       & (routes_df["destinations"] == destination_idx)]["path"].values)
                    for idx, path in enumerate(paths):
                        print(f'<route id="{origin_idx}_{destination_idx}_{idx}" edges="', file=rou)
                        print(path, file=rou)
                        print('" />',file=rou)
            print("</routes>", file=rou)

    def _get_detectors(self):

        paths_df = pd.read_csv(self.paths_csv_file_path)
        paths_list = [path.split(" ") for path in paths_df["path"].values]
        detectors_name = sorted(list(set([node for path in paths_list for node in path])))
        
        detectors_df = pd.DataFrame({"name": detectors_name})
        detectors_df.to_csv(self.detector_save_path, index=False)
        
        with open(self.det_xml_save_path, "w") as det:
            print("""<additional>""", file=det)
            for det_id in detectors_name:
                print(f"<inductionLoop id=\"{det_id}_det\" lane=\"{det_id}_0\" pos=\"-5\" file=\"NUL\" friendlyPos=\"True\"/>",
                      file=det)
            print("</additional>", file=det)
            
        return detectors_name

    ################################
    ######## SUMO CONTROL ##########
    ################################

    def start(self) -> None:
        """Starts the SUMO simulation with the specified configuration.

        Returns:
            None
        """

        sumo_cmd = [self.sumo_type,"--seed",
                    str(self.seed),
                    "--fcd-output",
                    self.sumo_fcd,
                    "-c", self.sumo_config_path]
        traci.start(sumo_cmd, label=self.sumo_id)
        self.sumo_connection = traci.getConnection(self.sumo_id)

    def stop(self) -> None:
        """Stops and closes the SUMO simulation.

        Returns:
            None
        """

        self.sumo_connection.close()

    def reset(self) -> dict:
        """ Resets the SUMO simulation to its initial state.

        Reads detector data.

        Returns:
            det_dict (dict[str, float]): dictionary with detector data.
        """

        det_dict = {name: None for name in self.detectors_name}
        for det_name in self.detectors_name:
            det_dict[det_name]  = self.sumo_connection.inductionloop.getIntervalVehicleNumber(f"{det_name}_det")

        self.sumo_connection.load(["--seed",
                                   str(self.seed),
                                   "--fcd-output",
                                   self.sumo_fcd,
                                   '-c',
                                   self.sumo_config_path])

        self.timestep = 0
        return det_dict

    ################################
    ######### SIMULATION ###########
    ################################

    def add_vehicle(self, act_dict: dict) -> None:
        """Adds a vehicle to the SUMO simulation environment with the specified route and parameters.

        Args:
            act_dict (dict): A dictionary containing key vehicle attributes.
        Returns:
            None
        """

        route_id = (
            self.route_id_cache.setdefault((
                act_dict[kc.AGENT_ORIGIN],
                act_dict[kc.AGENT_DESTINATION],
                act_dict[kc.ACTION]),
                f'{act_dict[kc.AGENT_ORIGIN]}_{act_dict[kc.AGENT_DESTINATION]}_{act_dict[kc.ACTION]}'))
        kind = act_dict[kc.AGENT_KIND]
        self.sumo_connection.vehicle.add(vehID=str(act_dict[kc.AGENT_ID]),
                                         routeID=route_id,
                                         depart=str(act_dict[kc.AGENT_START_TIME]),
                                         typeID=kind)

    def step(self) -> tuple:
        """Advances the SUMO simulation by one timestep and retrieves information
        about vehicle arrivals and detector data.

        Returns:
            self.timestep (int): The current simulation timestep.
            arrivals (list): List of vehicle IDs that arrived at their destinations during the current timestep.
        """
   
        arrivals = self.sumo_connection.simulation.getArrivedIDList()
        self.sumo_connection.simulationStep()
        self.timestep += 1
        
        return self.timestep, arrivals


import numpy as np
import random

from abc import ABC, abstractmethod

from routerl.keychain import Keychain as kc


class BaseLearningModel(ABC):
    """
    This is an abstract base class for the learning models used to model human learning and decision-making.\n
    Users can create their own learning models by inheriting from this class.
    """

    def __init__(self):
        pass

    @abstractmethod
    def act(self, state) -> None:
        """Method to select an action based on the current state and cost.

        Returns:
            None
        """
        pass

    @abstractmethod
    def learn(self, state, action, reward) -> None:
        """Method to learn the model based on the current state and cost.

        Arguments:
            state (Any): The current state of the environment.
            action (Any): The action to take.
            reward (Any): The reward received from the environment.
        Returns:
            None
        """

        pass


class Gawron(BaseLearningModel):
    """
    The Gawron learning model. This model is based on: `Gawron (1998) <https://kups.ub.uni-koeln.de/9257/>`_\n
    In summary, it iteratively shifts the cost expectations towards the received reward.\n
    For decision-making, calculates action utilities based on the ``beta`` parameter and cost expectations, and selects the action with the lowest utility.
    
    Args:
        params (dict): A dictionary containing model parameters.
        initial_knowledge (list or array): Initial knowledge of cost expectations.
    Attributes:
        beta (float): A parameter representing deviations in individual decision-making.
        alpha_zero (float): Agent's adaptation to new experiences.
        alpha_j (float): Weight for previous cost expectation (1 - ALPHA_ZERO).
        cost (np.ndarray): Agent's cost expectations for each option.
    """

    def __init__(self, params, initial_knowledge):
        super().__init__()

        # Extract beta with added randomness
        beta_randomness = params[kc.BETA_RANDOMNESS]
        self.beta = random.uniform(params[kc.BETA] - beta_randomness, params[kc.BETA] + beta_randomness)

        # Learning rate components
        self.alpha_zero = params[kc.ALPHA_ZERO]
        self.alpha_j = 1.0 - self.alpha_zero

        # Initialize cost array with initial knowledge
        self.cost = np.array(initial_knowledge, dtype=float)

    def act(self, state) -> int:
        """Selects an action based on the cost expectations.

        Args:
            state (Any): The current state of the environment (not used).
        Returns:
            action (int): The index of the selected action.
        """

        utilities = list(map(lambda x: np.exp(x * self.beta), self.cost))
        action =  utilities.index(min(utilities))
        return action   

    def learn(self, state, action, reward) -> None:
        """Updates the cost associated with the taken action based on the received reward.

        Args:
            state (string): The current state of the environment (not used).
            action (int): The action that was taken.
            reward (float): The reward received after taking the action.
        Returns:
            None
        """

        self.cost[action] = (self.alpha_j * self.cost[action]) + (self.alpha_zero * reward)


class Culo(BaseLearningModel):
    """
    The CUmulative LOgit learning model. This model is based on: `Li et al. (2024) <https://pubsonline.informs.org/doi/abs/10.1287/trsc.2023.0132/>`_.\n
    In summary, it updates its cost expectations by iteratively accumulating perceived rewards.\n
    For decision-making, calculates action utilities based on the ``beta`` parameter and cost expectations, and selects the action with the lowest utility.

    Args:
        params (dict): A dictionary containing model parameters.
        initial_knowledge (list or array): Initial knowledge of cost expectations.
    Attributes:
        beta (float): A parameter representing deviations in individual decision-making.
        alpha_zero (float): Agent's adaptation to new experiences.
        alpha_j (float): Weight for previous cost expectation (constant = 1).
        cost (np.ndarray): Agent's cost expectations for each option.
    """

    def __init__(self, params, initial_knowledge):
        super().__init__()

        # Extract beta with randomness
        beta_randomness = params[kc.BETA_RANDOMNESS]
        self.beta = random.uniform(params[kc.BETA] - beta_randomness, params[kc.BETA] + beta_randomness)

        # Learning rate components
        self.alpha_zero = params[kc.ALPHA_ZERO]
        self.alpha_j = 1

        # Initialize cost array with initial knowledge
        self.cost = np.array(initial_knowledge, dtype=float)

    def act(self, state) -> int:
        """Selects an action based on the cost expectations.

        Args:
            state (Any): The current state of the environment (not used).
        Returns:
            action (int): The index of the selected action.
        """

        utilities = list(map(lambda x: np.exp(x * self.beta), self.cost))
        action =  utilities.index(min(utilities))
        return action   

    def learn(self, state, action, reward) -> None:
        """Updates the cost associated with the taken action based on the received reward.

        Args:
            state (Any): The current state of the environment (not used).
            action (int): The action that was taken.
            reward (float): The reward received after taking the action.
        Returns:
            None
        """

        self.cost[action] = (self.alpha_j * self.cost[action]) + (self.alpha_zero * reward)


class WeightedAverage(BaseLearningModel):
    """
    Weighted Average learning model. Theory based on: `Cascetta (2009) <https://link.springer.com/book/10.1007/978-0-387-75857-2/>`_.\n
    In summary, the model uses the reward and a weighted average of the past cost expectations to update the current cost expectation.\n
    For decision-making, calculates action utilities based on the ``beta`` parameter and cost expectations, and selects the action with the lowest utility.
    

    Args:
        params (dict): A dictionary containing model parameters.
        initial_knowledge (list or array): Initial knowledge of cost expectations.
    Attributes:
        beta (float): A parameter representing deviations in individual decision-making.
        alpha_zero (float): Agent's adaptation to new experiences.
        alpha_j (float): Weight for previous cost expectation (1 - ALPHA_ZERO).
        remember (string): Memory size.
        cost (np.ndarray): Agent's cost expectations for each option.
        memory (list(list)): A list of lists containing the memory of each state.
    """

    def __init__(self, params, initial_knowledge):
        super().__init__()
        beta_randomness = params[kc.BETA_RANDOMNESS]
        self.beta = random.uniform(params[kc.BETA] - beta_randomness, params[kc.BETA] + beta_randomness)
        self.alpha_zero = params[kc.ALPHA_ZERO]
        self.alpha_j = 1.0 - self.alpha_zero
        self.remember = params[kc.REMEMBER]
        self.cost = np.array(initial_knowledge, dtype=float)
        self.memory = [list() for _ in range(len(initial_knowledge))]
        self.create_memory()

    def act(self, state) -> int:
        """Selects an action based on the cost expectations.

        Args:
            state (Any): The current state of the environment (not used).
        Returns:
            action (int): The index of the selected action.
        """

        utilities = list(map(lambda x: np.exp(x * self.beta), self.cost))
        action =  utilities.index(min(utilities))
        return action

    def learn(self, state, action, reward) -> None:
        """Updates the cost associated with the taken action based on the received reward.

        Args:
            state (Any): The current state of the environment (not used).
            action (int): The action that was taken.
            reward (float): The reward received after taking the action.
        Returns:
            None
        """

        # Drop the least relevant memory (end of list)
        del(self.memory[action][-1])
        # Insert the most recent expected cost at index 0
        self.memory[action].insert(0, self.cost[action])
        
        # Calculate the weights of the memory
        # The weights are proportional to item recency
        alpha_j_weights = [self.alpha_j / (memory_idx + 1) for memory_idx in range(self.remember)]
        # If remember=3 alpha_j=.5, then alpha_j_weights = [.5/1, .5/2, .5/3]. Now normalize alpha_j_weights.
        alpha_j_normalized = [a_j / sum(alpha_j_weights) for a_j in alpha_j_weights]
        
        # Calculate the weighted average of the memory
        c_hat = 0
        for memory_idx, a_j in enumerate(alpha_j_normalized):
            c_hat += a_j * self.memory[action][memory_idx]
            
        # Update the cost expectation of the action
        self.cost[action] = c_hat + (self.alpha_zero * reward)
        

    def create_memory(self) -> None:
        """
        Creates a memory of previous cost expectations.

        Returns:
            None
        """

        for i in range(len(self.cost)):
            for _ in range(self.remember):
                self.memory[i].append(self.cost[i])


from .agent import BaseAgent
from .agent import HumanAgent
from .agent import MachineAgent

from .agent_generation import generate_agents
from .agent_generation import generate_agent_data

from .simulator import SumoSimulator

from .observations import Observations
from .observations import PreviousAgentStart
from .observations import PreviousAgentStartPlusStartTime

from .environment import TrafficEnvironment


"""
Observation functions for RL agents (AVs).
"""

from gymnasium.spaces import Box
import numpy as np
from abc import ABC, abstractmethod
from typing import List, Dict, Any, Union

from routerl.keychain import Keychain as kc


class Observations(ABC):
    """Abstract base class for observation functions.

    Args:
        machine_agents_list (List[Any]): List of machine agents.
        human_agents_list (List[Any]): List of human agents.
    """

    def __init__(self, machine_agents_list: List[Any], human_agents_list: List[Any]) -> None:
        self.machine_agents_list = machine_agents_list
        self.human_agents_list = human_agents_list

    @abstractmethod
    def __call__(self, all_agents: List[Any]) -> Dict[str, Any]:
        pass

    @abstractmethod
    def observation_space(self) -> Dict[str, Box]:
        """Define the observation space for the observation function.

        Returns:
            Dict[str, Box]: A dictionary where keys are agent IDs and values are Gym spaces.
        """

        pass


class PreviousAgentStart(Observations):
    """Observes the number of agents with the same origin-destination and start time within a threshold.

    Args:
        machine_agents_list (List[Any]): List of machine agents.
        human_agents_list (List[Any]): List of human agents.
        simulation_params (Dict[str, Any]): Simulation parameters.
        agent_params (Dict[str, Any]): Agent parameters.
    Attributes:
        observations (List[Any]): List of observations.
    """

    def __init__(
        self,
        machine_agents_list: List[Any],
        human_agents_list: List[Any],
        simulation_params: Dict[str, Any],
        agent_params: Dict[str, Any],
        ) -> None:

        super().__init__(machine_agents_list, human_agents_list)
        self.simulation_params = simulation_params
        self.agent_params = agent_params
        self.observations = self.reset_observation()

    def __call__(self, all_agents: List[Any]) -> Dict[str, Any]:
        for machine in self.machine_agents_list:
            observation = np.zeros(self.simulation_params[kc.NUMBER_OF_PATHS], dtype=int)

            for agent in all_agents:
                if (machine.id != agent.id and
                    machine.origin == agent.origin and
                    machine.destination == agent.destination and
                    abs(machine.start_time - agent.start_time) < machine.observed_span):
                    
                    observation[agent.last_action] += 1

            self.observations[str(machine.id)] = observation.tolist()

        return self.observations

    def reset_observation(self) -> Dict[str, np.ndarray]:
        """Reset observations to the initial state.

        Returns:
            Dict[str, np.ndarray]: A dictionary of initial observations for all machine agents.
        """

        return {
            str(agent.id): np.zeros(self.simulation_params[kc.NUMBER_OF_PATHS], dtype=np.float32)
            for agent in self.machine_agents_list
        }

    def observation_space(self) -> Dict[str, Box]:
        """Define the observation space for each machine agent.

        Returns:
            Dict[str, Box]: A dictionary where keys are agent IDs and values are Gym spaces.
        """

        return {
            str(agent.id): Box(
                low=0,
                high=len(self.human_agents_list) + len(self.machine_agents_list),
                shape=(self.simulation_params[kc.NUMBER_OF_PATHS],),
                dtype=np.float32
            )
            for agent in self.machine_agents_list
        }

    def agent_observations(self, agent_id: str) -> np.ndarray:
        """Retrieve the observation for a specific agent.

        Args:
            agent_id (str): The ID of the agent.
        Returns:
            np.ndarray: The observation array for the specified agent.
        """

        return self.observations.get(agent_id, np.zeros(self.simulation_params[kc.NUMBER_OF_PATHS], dtype=np.float32))


class PreviousAgentStartPlusStartTime(Observations):
    """Observes the number of agents with the same origin-destination and start time within a threshold
    and includes the start of the specific agent as well.
    """

    def __init__(
        self,
        machine_agents_list: List[Any],
        human_agents_list: List[Any],
        simulation_params: Dict[str, Any],
        agent_params: Dict[str, Any],
    ) -> None:
        """Initialize the observation function.

        Args:
            machine_agents_list (List[Any]): List of machine agents.
            human_agents_list (List[Any]): List of human agents.
            simulation_params (Dict[str, Any]): Dictionary of simulation parameters.
            agent_params (Dict[str, Any]): Dictionary of agent parameters.
        Returns:
            None
        """

        super().__init__(machine_agents_list, human_agents_list)
        self.simulation_params = simulation_params
        self.agent_params = agent_params
        self.observations = self.reset_observation()
        self.agent_vectors = {}

    def __call__(self, all_agents: List[Any]) -> Dict[str, Any]:
        """Generate observations for all agents.

        Args:
            all_agents (List[Any]): List of all agents.
        Returns:
            Dict[str, Any]: A dictionary of observations keyed by agent IDs.
        """

        for machine in self.machine_agents_list:
            observation = np.zeros(self.simulation_params[kc.NUMBER_OF_PATHS], dtype=np.int32)

            for agent in all_agents:
                if (machine.id != agent.id and
                    machine.origin == agent.origin and
                    machine.destination == agent.destination and
                    machine.start_time > agent.start_time):
                    
                    observation[agent.last_action] += 1

            self.observations[str(machine.id)] = np.concatenate(
                [
                    np.array([machine.start_time], dtype=np.int32),  # Start time as scalar
                    observation  # Vector of observations
                ]
            )

        return self.observations

    def reset_observation(self) -> Dict[str, np.ndarray]:
        """Reset observations to the initial state.

        Returns:
            obs (Dict[str, np.ndarray]): A dictionary of initial observations for all machine agents.
        """

        # Initialize agent vectors as zero arrays
        self.agent_vectors = {
            agent: np.zeros(self.simulation_params[kc.NUMBER_OF_PATHS], dtype=np.int32)
            for agent in self.machine_agents_list
        }
        
        # Gather observations in a consistent format
        obs = {
            str(agent.id): np.concatenate(  # Combine start_time and vector into a single array
                [
                    np.array([agent.start_time], dtype=np.int32),  # Start time as scalar
                    self.agent_vectors[agent]  # Vector as array
                ]
            )
            for agent in self.machine_agents_list
        }

        return obs

    def observation_space(self) -> Dict[str, Box]:
        """
        Define the observation space for each machine agent.

        Returns:
            Dict[str, Box]: A dictionary where keys are agent IDs and values are Gym spaces.
        """

        total_size = 1 + self.simulation_params[kc.NUMBER_OF_PATHS]

        return {
            str(agent.id): Box(
                low=0,
                high=np.inf,
                shape=(total_size,),  # Combined size for start_time and vector
                dtype=np.float32
            )
            for agent in self.machine_agents_list
        }
    
    def agent_observations(self, agent_id: str) -> np.ndarray:
        """Retrieve the observation for a specific agent.

        Args:
            agent_id (str): The ID of the agent.
        Returns:
            np.ndarray: The observation array for the specified agent.
        """

        return self.observations.get(agent_id, np.zeros(self.simulation_params[kc.NUMBER_OF_PATHS], dtype=np.int32))


from .utils import confirm_env_variable
from .utils import df_to_prettytable
from .utils import get_params
from .utils import list_to_string
from .utils import make_dir
from .utils import show_progress
from .utils import show_progress_bar
from .utils import string_to_list
from .utils import remove_double_quotes
from .utils import resolve_param_dependencies
from .utils import running_average
from .utils import update_params


"""
Import all the necessary modules for the routerl package.

Some example imports:
    from routerl import Keychain
    from routerl import TrafficEnvironment
    from routerl import SumoSimulator
    from routerl import get_learning_model
    from routerl import plotter
    from routerl import Recorder
    
    from routerl.learning import Gawron
    from routerl.environment import HumanAgent
    from routerl.environment import PreviousAgentStart
"""

__version__ = "1.0.1"

from .keychain import Keychain

from .environment import (
    generate_agents,
    generate_agent_data,
    SumoSimulator,
    TrafficEnvironment
)

from .human_learning import (
    get_learning_model
)

from .services import (
    plotter,
    Plotter,
    Recorder
)

#TODO create_agents


class Keychain:

    """
    This is where we store our file paths, parameter access keys and other constants reliably
    When change needed, just fix it here.
    """

    ################ PARAMETER ACCESS KEYS ####################
    
    PARAMS_FILE = "params.json"
    PLOTTER_CONFIG_FILE = "plotter_config.json"

    AGENTS = "agent_parameters"
    ENVIRONMENT = "environment_parameters"
    SIMULATOR = "simulator_parameters"
    PLOTTER = "plotter_parameters"
    PATH_GEN = "path_generation_parameters"

    ### Agent parameters
    HUMAN_PARAMETERS = "human_parameters"
    MACHINE_PARAMETERS = "machine_parameters"
    # Common
    ACTION_SPACE_SIZE = "action_space_size"
    MODEL = "model"
    BEHAVIOR = "behavior"
    # Human
    BETA = "beta"
    BETA_RANDOMNESS = "beta_randomness"
    ALPHA_J = "alpha_j"
    ALPHA_ZERO = "alpha_zero"
    REMEMBER = "remember"
    # Machine
    OBSERVED_SPAN = "observed_span"
    OBSERVATION_TYPE = "observation_type"
    PREVIOUS_AGENTS = "previous_agents"
    PREVIOUS_AGENTS_PLUS_START_TIME = "previous_agents_plus_start_time"

    ### Environment
    NUMBER_OF_DAYS = "number_of_days"

    ### Simulator
    NETWORK_NAME = "network_name"
    SUMO_TYPE = "sumo_type"
    SIMULATION_TIMESTEPS = "simulation_timesteps"

    ### Plotter
    RECORDS_FOLDER = "records_folder"
    PLOTS_FOLDER = "plots_folder"
    PHASES = "phases"
    PHASE_NAMES = "phase_names"
    SMOOTH_BY = "smooth_by"

    ### Agent generation
    NUM_AGENTS = "num_agents"
    NEW_MACHINES_AFTER_MUTATION = "new_machines_after_mutation"

    ### Path generation
    NUMBER_OF_PATHS = "number_of_paths"
    ORIGINS = "origins"
    DESTINATIONS = "destinations"
    WEIGHT = "weight"
    NUM_SAMPLES = "num_samples"
    MAX_PATH_LENGTH = "max_path_length"
    ROUTE_UTILITY_COEFFS = "route_utility_coeffs"

    ###########################################################
    
    #################### CONSTANTS ############################

    NOT_AVAILABLE = "N/A"
    
    # Sumo
    ENV_VAR = "SUMO_HOME"
    
    # Common dataframe headers
    AGENT_ID = "id"
    AGENT_ORIGIN = "origin"
    AGENT_DESTINATION = "destination"
    AGENT_START_TIME = "start_time"
    AGENT_KIND = "kind"
    
    ACTION = "action"
    REWARD = "reward"
    COST = "cost"
    COST_TABLE = "cost_table"
    TRAVEL_TIME = "travel_time"
    FREE_FLOW_TIME = "free_flow_time"

    # Agent type encodings
    ALL = "All"
    TYPE_HUMAN = "Human"
    TYPE_MACHINE = "AV"

    # Behavior encodings
    SELFISH = "selfish"
    COMPETITIVE = "competitive"
    COLLABORATIVE = "collaborative"
    SOCIAL = "social"
    ALTRUISTIC = "altruistic"
    MALICIOUS = "malicious"
    BEHAVIORS = [SELFISH, COMPETITIVE, COLLABORATIVE, SOCIAL, ALTRUISTIC, MALICIOUS]

    # Model encodings
    GAWRON = "gawron"
    CULO = "culo"
    W_AVG = "w_avg"
    HUMAN_MODELS = [GAWRON, CULO, W_AVG]
    
    # Network names
    ARTERIAL = "arterial"
    COLOGNE = "cologne"
    CSOMOR = "csomor"
    GRID = "grid"
    INGOLSTADT = "ingolstadt"
    NGUYEN = "nguyen"
    ORTUZAR = "ortuzar"
    TWO_ROUTE_YIELD = "two_route_yield"
    MANHATTAN = "manhattan"
    NETWORK_NAMES = [ARTERIAL, COLOGNE, CSOMOR, GRID, INGOLSTADT, NGUYEN, ORTUZAR, TWO_ROUTE_YIELD, MANHATTAN]
    
    ###########################################################

    ####################### FILE PATHS ########################
    
    PATHS_CSV_FILE_NAME = "paths.csv"
    AGENTS_CSV_FILE_NAME = "agents.csv"
    
    NETWORK_FOLDER = "../networks/$net$/"
    CONNECTION_FILE_PATH = "../networks/$net$/$net$.con.xml"
    EDGE_FILE_PATH = "../networks/$net$/$net$.edg.xml"
    NOD_FILE_PATH = "../networks/$net$/$net$.nod.xml"
    SUMO_CONFIG_PATH = "../networks/$net$/$net$.sumocfg"
    ROU_FILE_PATH = "../networks/$net$/$net$.rou.xml"
    SUMO_FCD = "../networks/$net$/fcd.xml"
    ROUTE_XML_PATH = "../networks/$net$/route.rou.xml"
    DETECTORS_XML_PATH = "../networks/$net$/det.add.xml"
    DETECTORS_CSV_PATH = "../networks/$net$/detectors.csv"
    
    DEFAULT_ODS_PATH = "../networks/default_ods.json"
    
    EPISODES_LOGS_FOLDER = "episodes"
    DETECTOR_LOGS_FOLDER = "detector"
    LOSSES_LOG_FILE_NAME = "losses.txt"

    REWARDS_PLOT_FILE_NAME = "rewards.png"
    TRAVEL_TIMES_PLOT_FILE_NAME = "travel_times.png"
    TT_DIST_PLOT_FILE_NAME = "tt_dist.png"
    SIM_LENGTH_PLOT_FILE_NAME = "simulation_length.png"
    LOSSES_PLOT_FILE_NAME = "losses.png"
    ACTIONS_PLOT_FILE_NAME = "actions.png"
    ACTIONS_SHIFTS_PLOT_FILE_NAME = "actions_shifts.png"
    
    ###########################################################

    ###################### ZENODO DATA ########################

    ZENODO_RECORD_ID = 14866276

    ###########################################################


import shutil
import pytest
import subprocess
from pathlib import Path

TUTORIALS_DIR = Path("tutorials")
python_scripts = list(TUTORIALS_DIR.rglob("*.py"))

print(f"[DEBUG] Looking for Python scripts in {TUTORIALS_DIR.resolve()}")
print(f"[DEBUG] Found {len(python_scripts)} Python scripts.")

@pytest.fixture(scope="session", autouse=True)
def check_sumo_installed():
    sumo_executable = shutil.which("sumo")
    if sumo_executable is None:
        pytest.exit("[SUMO ERROR] SUMO is not installed or not in PATH.")

@pytest.mark.parametrize("script_path", python_scripts)
def test_python_script_execution(script_path):
    try:
        script_filename = script_path.name
        result = subprocess.run(
            ["python", script_filename], capture_output=True, text=True, check=True, cwd=script_path.parent
        )
        print(f"[DEBUG] Successfully executed {script_path}")
    except subprocess.CalledProcessError as e:
        pytest.fail(f"[FAIL] Script {script_path} failed to execute: {e.stderr}")


"""
PettingZoo environment for optimal route choice using SUMO simulator.

"""

import os

from copy import copy
from copy import deepcopy as dc
from gymnasium.spaces import Discrete

import functools
import logging
import numpy as np
import pandas as pd
import random
import threading

from routerl.environment import generate_agents
from routerl.environment import SumoSimulator
from routerl.environment import MachineAgent
from routerl.environment import PreviousAgentStart, PreviousAgentStartPlusStartTime, Observations
from routerl.keychain import Keychain as kc
from routerl.services import plotter
from routerl.services import Recorder
from routerl.utilities import get_params

from pettingzoo.utils.env import AECEnv
from pettingzoo.utils import agent_selector

logger = logging.getLogger()
logger.setLevel(logging.WARNING)


class TrafficEnvironment(AECEnv):
    """
    A PettingZoo AECEnv interface for optimal route choice using SUMO simulator. 
    This environment is designed for the training of human agents (rational decision-makers) 
    and machine agents (reinforcement learning agents).
    
    See `SUMO <https://sumo.dlr.de/docs/>`_ for details on SUMO. \n
    See `PettingZoo <https://pettingzoo.farama.org/>`_ for details on PettingZoo. 
    
    .. note::
        Users can configure the experiment with keyword arguments, see the structure below. 
        Moreover, users can provide custom demand data in ``training_records/agents.csv``.
        You can refer to the structure of such a file `here <https://github.com/COeXISTENCE-PROJECT/RouteRL/blob/main/docs/_static/agents_example.csv>`_.

    Args:
        seed (int, optional): 
            Random seed for reproducibility. Defaults to ``23423``.
        create_agents (bool, optional):
            Whether to create agent data. Defaults to ``True``.
        create_paths (bool, optional):
            Whether to generate paths. Defaults to ``True``.
        **kwargs (dict, optional): 
            User-defined parameter overrides. These override default values 
            from ``params.json`` and allow experiment configuration.
            

    Keyword arguments (see the usage below):
    
        - agent_parameters (dict, optional):
            Agent settings.
            
            - num_agents (int, default=100):
                Total number of agents.
            
            - new_machines_after_mutation (int, default=25):
                Number of humans converted to machines.
            
            - machine_parameters (**dict**):
                Machine agent settings.
                
                - behavior (str, default="selfish"):
                    Route choice behavior.
                    Options: ``selfish``, ``social``, ``altruistic``, ``malicious``, ``competitive``, ``collaborative``.
                    
                - observed_span (int, default=300):
                    Time window considered for observations.
                    
                - observation_type (str, default="previous_agents_plus_start_time"):
                    Type of observation.
                    Options: ``previous_agents``, ``previous_agents_plus_start_time``.

            - human_parameters (**dict**): 
                Human agent settings.
                
                - model (str, default="culo"):
                    Decision-making model (options: ``gawron``, ``culo``, ``w_avg``).
                    
                - alpha_j (float, default=0.5):
                    Cost expectation coefficient (0-1 range).
                    
                - alpha_zero (float, default=0.5):
                    Sensitivity to new experiences.
                    
                - beta (float, default=-1.5):
                    Decision randomness parameter.
                    
                - beta_randomness (float, default=0.1):
                    Variability in ``beta`` among the human population.
                    
                - remember (int, default=3):
                    Number of past experiences retained.

        - environment_parameters (dict, optional):
            Environment settings.
            
            - number_of_days (int, default=1):
                Number of days in the scenario.

        - simulator_parameters (dict, optional): 
            **SUMO simulator settings.**
            
            - network_name (str, default="csomor"):
                Network name (e.g., ``arterial``, ``cologne``, ``grid``).
            
            - simulation_timesteps (int, default=180):
                Total simulation time in seconds.
            
            - sumo_type (str, default="sumo"):
                SUMO execution mode (``sumo`` or ``sumo-gui``).

        - path_generation_parameters (dict, optional):
            Path generation settings.
            
            - number_of_paths (int, default=3):
                Number of routes per OD.
                
            - beta (float, default=-3.0):
                Sensitivity to travel time in path generation.
                
            - weight (str, default="time"):
                Optimization criterion.
                
            - num_samples (int, default=100):
                Number of samples for path generation.
                
            - origins (str | list[str], default="default"):
                Origin points from the network. (e.g., ``["-25166682#0", "-4936412"]``)
                
            - destinations (str | list[str], default="default"):
                Destination points from the network. (e.g., ``["-115604057#1", "-279952229#4"]``)

        - plotter_parameters (dict, optional): 
            Plotting & logging settings.
            
            - records_folder (str, default="training_records"):
                Directory for training records.
                
            - plots_folder (str, default="plots"):
                Directory for plots.
                
            - smooth_by (int, default=50): 
                Smoothing parameter for plots.
                
            - phases (list[int], default=[0, 100]):
                X-axis positions for phase markers.
                
            - phase_names (list[str], default=["Human learning", "Mutation - Machine learning"]):
                Phase names for labeling phase markers.
    
    Usage:
        
        .. rubric:: Case 1
        
        .. code-block:: text
        
            % Your file structure in the beginning
            project_directory/
            |-- your_script.py
            
        .. code-block:: python
        
            >>> # Environment initialization
            ... env = TrafficEnvironment(
            ...     seed=42,
            ...     agent_parameters={
            ...         "num_agents": 5, 
            ...         "new_machines_after_mutation": 1, 
            ...         "machine_parameters": {
            ...             "behavior": "selfish"
            ...             }},
            ...     simulator_parameters={"sumo_type": "sumo-gui"},
            ...     path_generation_parameters={"number_of_paths": 2}
            ... )
            
        .. code-block:: text
        
            % File structure after the initialization:
            project_directory/
            |-- your_script.py
            |-- training_records/
            |   |-- agents.csv
            |   |-- paths.csv
            |   |-- detector/
            |   |   |--             % to be populated during simulation
            |   |-- episodes/
            |   |   |--             % to be populated during simulation
            |-- plots/
            |   |-- 0_0.png 
            |   |-- ...             % visuals of generated paths for each OD
            |   |-- ...             % to be populated after the experiment
            
        .. raw:: html

            <hr style="border:1px solid #ccc; margin: 20px 0;">
        
        .. rubric:: Case 2
        
        .. code-block:: text
        
            % Your file structure in the beginning
            project_directory/
            |-- your_script.py
            |-- training_records/
            |   |-- agents.csv      % your custom demand, conforming to the structure
        
        .. warning::
            Demand data in ``agents.csv`` should be aligned with the specified 
            experiment settings (e.g., number of agents, number of origins and destinations, etc.).
            
        .. code-block:: python
        
            >>> env = TrafficEnvironment(
            ...     create_agents=False, # Environment will use your agent data
            ...     agent_parameters={
            ...         "new_machines_after_mutation": 10, 
            ...         "machine_parameters": {
            ...             "behavior": "selfish"
            ...             }},
            ...     simulator_parameters={"network_name": "arterial"},
            ...     path_generation_parameters={"number_of_paths": 3}
            ... )
            
        .. code-block:: text
        
            % File structure after the initialization:
            project_directory/
            |-- your_script.py
            |-- training_records/
            |   |-- agents.csv      % stays the same, used for agent generation
            |   |-- paths.csv
            |   |-- detector/
            |   |   |--             % to be populated during simulation
            |   |-- episodes/
            |   |   |--             % to be populated during simulation
            |-- plots/
            |   |-- 0_0.png 
            |   |-- ...             % visuals of generated paths for each OD
            |   |-- ...             % to be populated after the experiment
            
        .. warning::
            Same approach does not translate to path generation.\n
            ``paths.csv`` is mainly used for visualization purposes. 
            For SUMO to operate correctly, a ``route.rou.xml`` 
            should be generated inside the ``routerl/networks/<net_name>/`` folder.\n
            It is advised to generate paths in each experiment providing a random seed,
            or set ``create_paths=False`` only when above criteria is met.

    Attributes:
        day (int): Current day index in the simulation.
        human_learning (bool): Whether human agents are learning.
        number_of_days (int): Number of days to simulate.
        action_space_size (int): Size of the action space.
        recorder (Recorder): Object for recording simulation data.
        simulator (SumoSimulator): SUMO simulator instance.
        all_agents (list): List of all agent objects.
        machine_agents (list): List of all machine agent objects.
        human_agents (list): List of all human agent objects.
    """
    
    metadata = {
        "render_modes": ["human"],
        "name": "TrafficEnvironment",
    }

    def __init__(self,
                 seed: int = 23423,
                 create_agents: bool = True,
                 create_paths: bool = True,
                 **kwargs) -> None:

        super().__init__()
        self.render_mode = None

        # Read default parameters, update w #TODO kwargs
        params_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), kc.PARAMS_FILE)
        params = get_params(params_path, resolve=True, update=kwargs)

        self.environment_params = params[kc.ENVIRONMENT]
        self.simulation_params = params[kc.SIMULATOR]
        self.agent_params = params[kc.AGENTS]
        self.plotter_params = params[kc.PLOTTER]
        self.path_gen_params = params[kc.PATH_GEN] if create_paths else None

        self.travel_times_list = []
        self.day = 0
        self.human_learning = True
        self.machine_same_start_time = []
        self.actions_timestep = []

        self.number_of_days = self.environment_params[kc.NUMBER_OF_DAYS]
        self.action_space_size = self.environment_params[kc.ACTION_SPACE_SIZE]
        self._set_seed(seed)

        self.recorder = Recorder(self.plotter_params)
        self.simulator = SumoSimulator(self.simulation_params, self.path_gen_params, seed)

        self.all_agents = generate_agents(self.agent_params, self.get_free_flow_times(), create_agents, seed)
        self.machine_agents = [agent for agent in self.all_agents if agent.kind == kc.TYPE_MACHINE]
        self.human_agents = [agent for agent in self.all_agents if agent.kind == kc.TYPE_HUMAN]
        self.possible_agents = list()

        if len(self.machine_agents):
            self._initialize_machine_agents()
        if not self.human_agents:
            self.human_learning = False
        logging.info(f"There are {len(self.human_agents)} human and {len(self.machine_agents)} machine agents.")

        self.episode_actions = dict()

    def __str__(self):
        message = f"TrafficEnvironment with {len(self.all_agents)} agents.\
            \n{len(self.machine_agents)} machines and {len(self.human_agents)} humans.\
            \nMachines: {sorted(self.machine_agents, key=lambda agent: agent.id)}\
            \nHumans: {sorted(self.human_agents, key=lambda agent: agent.id)}"
        return message

    def _set_seed(self, seed: int) -> None:
        random.seed(seed)
        np.random.seed(seed)
        self.seed = seed
        logging.info(f"Seed set to {seed}.")

    def _initialize_machine_agents(self) -> None:

        ## Sort machine agents based on their start_time
        sorted_machine_agents = sorted(self.machine_agents, key=lambda agent: agent.start_time)
        self.possible_agents = [str(agent.id) for agent in sorted_machine_agents]
        self.n_agents = len(self.possible_agents)

        self.agent_name_mapping = dict(
            zip(self.possible_agents, list(range(len(self.possible_agents))))
        )

        ## Initialize the observation object
        self.observation_obj = self.get_observation_function()

        self._observation_spaces = self.observation_obj.observation_space()

        self._action_spaces = {
            agent: Discrete(self.simulation_params[kc.NUMBER_OF_PATHS]) for agent in self.possible_agents
        }

        logging.info("\nMachine's observation space is: %s ", self._observation_spaces)
        logging.info("Machine's action space is: %s", self._action_spaces)

    ################################
    ######## Control methods #######
    ################################

    def start(self) -> None:
        """Start the connection with SUMO.
        
        Returns:
            None
        """

        self.simulator.start()

    def reset(self, seed: int = None, options: dict = None) -> tuple:
        """Resets the environment.
        
        Args:
            seed (int, optional): Seed for random number generation. Defaults to None.
            options (dict, optional): Additional options for resetting the environment. Defaults to None.
            
        Returns:
            observations (dict): observations.
            infos (dict): dictionary of information for the agents.
        """

        self.episode_actions = dict()
        self.simulator.reset()
        self.agents = copy(self.possible_agents)
        self.terminations = {agent: False for agent in self.possible_agents}
        self.truncations = {agent: False for agent in self.possible_agents}
        self._cumulative_rewards = {agent: 0 for agent in self.possible_agents}
        self.infos = {agent: {} for agent in self.possible_agents}
        self.rewards = {agent: 0 for agent in self.possible_agents}
        self.rewards_humans = {agent.id: 0 for agent in self.human_agents}
        self.travel_times_list = []

        if len(self.machine_agents) > 0:
            self._agent_selector = agent_selector(self.possible_agents)
            self.agent_selection = self._agent_selector.next()
            self.observations = self.observation_obj.reset_observation()
        else:
            self.observations = {}

        infos = {a: {} for a in self.possible_agents}

        return self.observations, infos

    def step(self, machine_action: int = None) -> None:
        """Step method.

        Takes an action for the current agent (specified by `agent_selection`) and updates
        various parameters including rewards, cumulative rewards, terminations, truncations,
        infos, and agent_selection. Also updates any internal state used by `observe()`.

        Args:
            machine_action (int, optional):
                The action to be taken by the machine agent. Defaults to None.
            
        Returns:
            None
        """

        # If there are machines in the system
        if self.possible_agents:
            if (self.terminations[self.agent_selection]
                    or self.truncations[self.agent_selection]):
                # handles stepping an agent which is already dead
                # accepts a None action for the one agent, and moves the agent_selection to
                # the next dead agent,  or if there are no more dead agents, to the next live agent
                self._was_dead_step(machine_action)
                return

            agent = self.agent_selection

            # The cumulative reward of the last agent must be 0
            self._cumulative_rewards[agent] = 0
            self.simulation_loop(machine_action, agent)

            # Collect rewards if it is the last agent to act
            if self._agent_selector.is_last():
                # Increase day number
                self.day += 1

                # Calculate the rewards
                self._assign_rewards()

                # The episode ends when we complete episode_length days
                self.truncations = {agent: not (self.day % self.number_of_days) for agent in self.agents}
                self.terminations = {agent: not (self.day % self.number_of_days) for agent in self.agents}
                self.infos = {agent: {} for agent in self.agents}
                self.observations = self.observation_obj(self.all_agents)
                self._reset_episode()
            else:
                # no rewards are allocated until all players give an action
                self._clear_rewards()
                self.agent_selection = self._agent_selector.next()

            # Adds .rewards to ._cumulative_rewards
            self._accumulate_rewards()

        # If there are only humans in the system
        else:
            self.simulation_loop(machine_action=0, machine_id=0)
            self.day = self.day + 1
            self._assign_rewards()
            self._reset_episode()

    def close(self) -> None:
        """Not implemented.

        Returns:
            None
        """
        pass
    
    def stop_simulation(self) -> None:
        """End the simulation.

        Returns:
            None
        """

        self.simulator.stop()

    def observe(self, agent: str) -> np.ndarray:
        """Retrieve the observations for a specific agent.

        Args:
            agent (str): The identifier for the agent whose observations are to be retrieved.
            
        Returns:
            self.observation_obj.agent_observations(agent) (np.ndarray): The observations for the specified agent.
        """

        return self.observation_obj.agent_observations(agent)

    #########################
    ### Mutation function ###
    #########################

    def mutation(self, disable_human_learning: bool = True) -> None:
        """Perform mutation by converting selected human agents into machine agents.

        This method identifies a subset of human agents that start after the 25th percentile of start times of
        other vehicles, removes a specified number of these agents, and replaces them with machine agents.

        Args:
            disable_human_learning (bool, optional): Boolean flag to disable human agents.
            
        Returns:
            None
            
        Raises:
            ValueError: If there are insufficient human agents available for mutation.
        """

        logging.info("Mutation is about to happen!\n")
        logging.info("There were %s human agents.\n", len(self.human_agents))

        # Mutate to a human that starts after the 25% of the rest of the vehicles
        #start_times = [human.start_time for human in self.human_agents]
        #percentile_25 = np.percentile(start_times, 25)

        #filtered_human_agents = [human for human in self.human_agents if human.start_time > percentile_25]
        filtered_human_agents = [human for human in self.human_agents]

        number_of_machines_to_be_added = self.agent_params[kc.NEW_MACHINES_AFTER_MUTATION]
        random_humans_deleted = []

        if len(filtered_human_agents) < number_of_machines_to_be_added:
            raise ValueError(
                f"Insufficient human agents for mutation. Required: {number_of_machines_to_be_added}, "
                f"Available: {len(filtered_human_agents)}.\n"
                f"Decrease the number of machines to be added after the mutation.\n"
            )

        for _ in range(0, number_of_machines_to_be_added):
            random_human = random.choice(filtered_human_agents)

            self.human_agents.remove(random_human)
            filtered_human_agents.remove(random_human)

            random_humans_deleted.append(random_human)
            self.machine_agents.append(MachineAgent(random_human.id,
                                                    random_human.start_time,
                                                    random_human.origin,
                                                    random_human.destination,
                                                    self.agent_params[kc.MACHINE_PARAMETERS],
                                                    self.action_space_size))
            self.possible_agents.append(str(random_human.id))

        self.n_agents = len(self.possible_agents)
        self.all_agents = self.machine_agents + self.human_agents

        if disable_human_learning:  self.human_learning = False

        logging.info(f"Now there are {len(self.human_agents)} human agents.")

        self._initialize_machine_agents()

    #########################
    ##### Help functions ####
    #########################

    def get_observation(self) -> tuple:
        """Retrieve the current observation from the simulator.

        This method returns the current timestep of the simulation and the values of the episode actions.

        Returns:
            tuple: A tuple containing the current timestep and the episode actions.
        """

        return self.simulator.timestep, self.episode_actions.values()

    def _help_step(self, actions: list[tuple]) -> dict:

        for agent, action in actions:
            action_dict = {kc.AGENT_ID: agent.id,
                           kc.AGENT_KIND: agent.kind,
                           kc.ACTION: action,
                           kc.AGENT_ORIGIN: agent.origin,
                           kc.AGENT_DESTINATION: agent.destination,
                           kc.AGENT_START_TIME: agent.start_time}
            self.simulator.add_vehicle(action_dict)
            self.episode_actions[agent.id] = action_dict
        timestep, arrivals = self.simulator.step()

        travel_times = dict()
        for veh_id in arrivals:
            agent_id = int(veh_id)
            travel_times[agent_id] = ({kc.TRAVEL_TIME:
                                           (timestep - self.episode_actions[agent_id][kc.AGENT_START_TIME]) / 60.0})
            travel_times[agent_id].update(self.episode_actions[agent_id])

        return travel_times.values()

    def _reset_episode(self) -> None:

        detectors_dict = self.simulator.reset()

        if self.possible_agents:
            self._agent_selector = agent_selector(self.possible_agents)
            self.agent_selection = self._agent_selector.next()

        recording_task = threading.Thread(target=self._record, args=(self.day,
                                                                     self.travel_times_list,
                                                                     self.all_agents,
                                                                     detectors_dict))
        recording_task.start()

        self.travel_times_list = []
        self.episode_actions = dict()

    def _assign_rewards(self) -> None:

        for agent in self.all_agents:
            reward = agent.get_reward(self.travel_times_list)

            # Add the reward in the travel_times_list
            for agent_entry in self.travel_times_list:
                if agent.id == agent_entry[kc.AGENT_ID]:
                    self.travel_times_list.remove(agent_entry)
                    agent_entry[kc.REWARD] = reward
                    self.travel_times_list.append(agent_entry)

            # Save machine's rewards based on PettingZoo standards
            if agent.kind == 'AV':
                self.rewards[str(agent.id)] = reward

            # Human learning
            elif self.human_learning:
                agent.learn(agent.last_action, self.travel_times_list)

    ###########################
    ##### Simulation loop #####
    ###########################

    def simulation_loop(self, machine_action: int, machine_id: int) -> None:
        """This function contains the integration of the agent's actions to SUMO.

        We iterate through all the time steps of the simulation.
        For each timestep there are none, one or more than one agents type (humans, machines) that start.
        If more than one machine agents have the same start time, we break from this function because
        we need to take the agent's action from the STEP function.

        Args:
            machine_action (int): The id of the machine agent whose action is to be performed.
            machine_id (int): The id of the machine agent whose action is to be performed.
            
        Returns:
            None
        """

        agent_action = False
        while (
                self.simulator.timestep < self.simulation_params[kc.SIMULATION_TIMESTEPS]
                or len(self.travel_times_list) < len(self.all_agents)
        ):

            # If there are more than one machines with the same start time
            # the humans should act once
            if not self.actions_timestep:
                for human in self.human_agents:
                    if human.start_time == self.simulator.timestep:
                        action = human.act(0)
                        human.last_action = action
                        self.actions_timestep.append((human, action))

            for machine in self.machine_agents:
                if machine.start_time == self.simulator.timestep:

                    # In case there are machine agents that have the same start time, but it's not their turn
                    if str(machine.id) != machine_id:

                        # If some machines have the same start time, and they haven't acted yet
                        if (
                                (machine not in self.machine_same_start_time)
                                and not any(machine == item[0] for item in self.actions_timestep)
                        ):
                            self.machine_same_start_time.append(machine)
                        continue
                    else:
                        # Machine acting
                        machine.last_action = machine_action
                        self.actions_timestep.append((machine, machine_action))

                        # The machine acted should be deleted from the self.machine_same_start_time list
                        if machine in self.machine_same_start_time:
                            self.machine_same_start_time.remove(machine)

                        # If the machine isn't the last agent to act then we need to step again for the next agent
                        if not self._agent_selector.is_last():
                            agent_action = True

            # If all machines that have start time as the simulator timestep acted
            if not self.machine_same_start_time:
                travel_times = self._help_step(self.actions_timestep)

                for agent_dict in travel_times:
                    self.travel_times_list.append(agent_dict)

                self.actions_timestep = []
                self.machine_same_start_time = []

            # If the machine agent that had turn acted
            if agent_action:
                agent_action = False
                break

    ###########################
    ##### Free flow times #####
    ###########################

    def get_free_flow_times(self) -> dict:
        """Retrieve free flow times for all origin-destination pairs from the simulator paths data.

        Returns:
            ff_dict (dict): A dictionary where keys are tuples of origin and destination,
                            and values are lists of free flow times.
        """

        paths_df = pd.read_csv(self.simulator.paths_csv_file_path)
        origins = paths_df[kc.ORIGINS].unique()
        destinations = paths_df[kc.DESTINATIONS].unique()
        ff_dict = {(o, d): list() for o in origins for d in destinations}

        for _, row in paths_df.iterrows():
            ff_dict[(row[kc.ORIGINS], row[kc.DESTINATIONS])].append(row[kc.FREE_FLOW_TIME])

        return ff_dict

    ############################
    ##### Disc operations ######
    ############################

    def _record(self, episode: int, ep_observations: dict, agents: list, detectors_dict: dict) -> None:

        dc_episode, dc_ep_observations, dc_agents, dc_detectors = dc(episode), dc(ep_observations), dc(agents), dc(detectors_dict)
        zero_space = [0] * self.action_space_size

        cost_tables = [
            {
                kc.AGENT_ID: agent.id,
                kc.COST_TABLE: getattr(agent.model, 'cost', zero_space) if hasattr(agent, 'model') else zero_space
            }
            for agent in dc_agents
        ]

        self.recorder.record(dc_episode, dc_ep_observations, cost_tables, dc_detectors)

    def plot_results(self) -> None:
        """Method that plot the results of the simulation.

        Returns:
            None
        """

        plotter(self.plotter_params)

    ############################
    ### PettingZoo functions ###
    ############################

    def render(self) -> None:
        pass

    @functools.lru_cache(maxsize=None)
    def observation_space(self, agent: str):
        """Method that returns the observation space of the agent.

        Args:
            agent (str): The agent name.
        Returns:
            self._observation_spaces[agent] (Any): The observation space of the agent.
        """

        return self._observation_spaces[agent]

    @functools.lru_cache(maxsize=None)
    def action_space(self, agent: str):
        """Method that returns the action space of the agent.

        Args:
            agent (str): The agent name.
        Returns:
            self._action_spaces[agent] (Any): The action space of the agent.
        """

        return self._action_spaces[agent]

    #####################################################
    ### Decide on the observation function to be used ###
    #####################################################

    def get_observation_function(self) -> Observations:
        """Returns an observation object based on the provided parameters.

        Returns:
            Observations: An observation object.
        Raises:
            ValueError: If model is unknown.
        """

        params = self.agent_params[kc.MACHINE_PARAMETERS]
        observation_type = params[kc.OBSERVATION_TYPE]
        if observation_type == kc.PREVIOUS_AGENTS_PLUS_START_TIME:
            return PreviousAgentStartPlusStartTime(self.machine_agents,
                                                   self.human_agents,
                                                   self.simulation_params,
                                                   self.agent_params)
        elif observation_type == kc.PREVIOUS_AGENTS:
            return PreviousAgentStart(self.machine_agents,
                                      self.human_agents,
                                      self.simulation_params,
                                      self.agent_params)
        else:
            raise ValueError('[MODEL INVALID] Unrecognized observation type: ' + observation_type)



import json
import logging
import matplotlib.pyplot as plt
import numpy as np
import os
import pandas as pd
import seaborn as sns
import warnings

from collections import Counter
from matplotlib import rcParams
from statistics import mean, variance
from typing import Union

from routerl.keychain import Keychain as kc
from routerl.utilities import make_dir
from routerl.utilities import running_average
from routerl.utilities import get_params

warnings.filterwarnings("ignore")
logger = logging.getLogger()
logger.setLevel(logging.WARNING)


class Plotter:
    """Plot the results of the training

    Args:
        params (list): 
            Plotter parameters as specified `here <https://coexistence-project.github.io/RouteRL/documentation/pz_env.html#>`_.
            
    Methods:
        plot: plot the results
        visualise_mean_reward: visualise the mean reward
        visualise_mean_travel_times: visualise the mean travel times
        travel_tt_distributions: visualise the travel time distributions
        visualise_actions: visualise the actions
        visualise_action_shifts: visualise the action shifts
        visualise_sim_length: visualise the simulation length
        visualise_losses: visualise the losses
        
    Note:
        Plotter styling configuration is stored in `routerl/services/plotter_config.json` file.
    """

    def __init__(self, params):
        self.label_size = 12
        self.tick_label_size = 12
        self.line_width = 3
        self.title_size = 16
        self.default_width = 15
        self.default_height = 15
        self.legend_font_size = 12
        self.default_num_columns = 3
        self.multimode_width = 3
        self.multimode_height = 3
        self.linestyles = ['-']
        self.colors = []

        self.params = params
        self.phases = params[kc.PHASES]
        self.phase_names = params[kc.PHASE_NAMES]
        self.smooth_by = params[kc.SMOOTH_BY]
        self.records_folder = params[kc.RECORDS_FOLDER]

        with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), kc.PLOTTER_CONFIG_FILE), 'r') as file:
            config = json.load(file)
        for key, value in config.items():
            setattr(self, key, value)
        self.phase_colors = list(reversed(self.colors))
        rcParams['font.family'] = self.font_family
        
        make_dir(self.records_folder)
        self.episodes_folder = make_dir([self.records_folder, kc.EPISODES_LOGS_FOLDER])
        self.loss_file_path = make_dir(self.records_folder, kc.LOSSES_LOG_FILE_NAME)

        self.saved_episodes = list()

    ################################
    ######## VISUALIZE ALL #########
    ################################

    def plot(self) -> None:
        """Plot the results of the training

        Returns:
            None
        """

        self.saved_episodes = self._get_episodes()
        self.visualize_mean_rewards()
        self.visualize_mean_travel_times()
        self.visualize_tt_distributions()
        self.visualize_actions()
        self.visualize_action_shifts()
        self.visualize_sim_length()
        self.visualize_losses()

    def _get_episodes(self) -> list[int]:
        """Get the episodes data

        Returns:
            sorted_episodes (list[int]): the sorted episodes data
        Raises:
            FileNotFoundError: If the episodes folder does not exist
        """

        eps = list()
        if os.path.exists(self.episodes_folder):
            for file in os.listdir(self.episodes_folder):
                episode = int(file.split('ep')[1].split('.csv')[0])
                eps.append(episode)
        else:
            raise FileNotFoundError(f"Episodes folder does not exist!")
        return sorted(eps)

    ################################
    ########### REWARDS ############
    ################################

    def visualize_mean_rewards(self) -> None:
        """Visualise the mean rewards

        Returns:
            None
        """

        save_to = make_dir(self.params[kc.PLOTS_FOLDER], kc.REWARDS_PLOT_FILE_NAME)
        all_mean_rewards = self._retrieve_data_per_kind(kc.REWARD, transform='mean')

        plt.figure(figsize=(self.default_width, self.default_height), layout='tight')

        for idx, (kind, ep_reward_dict) in enumerate(all_mean_rewards.items()):
            episodes = list(ep_reward_dict.keys())
            rewards = list(ep_reward_dict.values())
            smoothed_rewards = running_average(rewards, last_n=self.smooth_by)
            plt.plot(episodes, smoothed_rewards, color=self.colors[idx], label=kind, linewidth=self.line_width)

        for phase_idx, phase in enumerate(self.phases):
            color = self.phase_colors[phase_idx % len(self.phase_colors)]
            plt.axvline(x=phase,
                        label=self.phase_names[phase_idx],
                        linestyle='--',
                        color=color,
                        linewidth=self.line_width)

        plt.xticks(fontsize=self.tick_label_size)
        plt.yticks(fontsize=self.tick_label_size)
        plt.xlabel('Episode', fontsize=self.label_size)
        plt.ylabel('Mean Reward', fontsize=self.label_size)
        plt.grid(True, axis='y')
        plt.title('Mean Rewards Over Episodes', fontsize=self.title_size, fontweight='bold')
        plt.legend(fontsize=self.legend_font_size)

        plt.savefig(save_to)
        plt.close()
        logging.info(f"[SUCCESS] Rewards are saved to {save_to}")

    ################################
    ######### TRAVEL TIMES #########
    ################################

    def visualize_mean_travel_times(self) -> None:
        """Visualise the mean travel times

        Returns:
            None
        """

        save_to = make_dir(self.params[kc.PLOTS_FOLDER], kc.TRAVEL_TIMES_PLOT_FILE_NAME)
        all_mean_tt = self._retrieve_data_per_kind(kc.TRAVEL_TIME, transform='mean')

        plt.figure(figsize=(self.default_width, self.default_height), layout='tight')

        for idx, (kind, ep_tt_dict) in enumerate(all_mean_tt.items()):
            episodes = list(ep_tt_dict.keys())
            tts = list(ep_tt_dict.values())
            smoothed_tts = running_average(tts, last_n=self.smooth_by)
            plt.plot(episodes, smoothed_tts, color=self.colors[idx], label=kind, linewidth=self.line_width)

        for phase_idx, phase in enumerate(self.phases):
            color = self.phase_colors[phase_idx % len(self.phase_colors)]
            plt.axvline(x=phase,
                        label=self.phase_names[phase_idx],
                        linestyle='--',
                        color=color,
                        linewidth=self.line_width)

        plt.xticks(fontsize=self.tick_label_size)
        plt.yticks(fontsize=self.tick_label_size)
        plt.xlabel('Episode', fontsize=self.label_size)
        plt.ylabel('Mean Travel Time', fontsize=self.label_size)
        plt.grid(True, axis='y')
        plt.title('Mean Travel Times Over Episodes', fontsize=self.title_size, fontweight='bold')
        plt.legend(fontsize=self.legend_font_size)

        plt.savefig(save_to)
        plt.close()
        logging.info(f"[SUCCESS] Travel times are saved to {save_to}")

    ################################
    ### TRAVEL TIME DISTRIBUTIONS ##
    ################################

    def visualize_tt_distributions(self) -> None:
        """Visualise the travel time distributions

        Returns:
            None
        """
        save_to = make_dir(self.params[kc.PLOTS_FOLDER], kc.TT_DIST_PLOT_FILE_NAME)
    
        num_rows, num_cols = 2, 2
        fig, axes = plt.subplots(num_rows,
                                 num_cols,
                                 figsize=(self.multimode_width * num_cols,
                                          self.multimode_height * num_rows))
        fig.tight_layout(pad=5.0)

        if num_rows > 1:
            axes = axes.flatten()   # Flatten axes
        if not hasattr(axes, '__getitem__'):
            axes = np.array([axes])  # If only one subplot

        # Plot mean travel times for each OD
        mean_tt_od = self._retrieve_data_per_od(kc.TRAVEL_TIME, transform='mean')
        sorted_keys = sorted(mean_tt_od.keys())
        for idx, od in enumerate(sorted_keys):
            episodes = list(mean_tt_od[od].keys())
            mean_tt = list(mean_tt_od[od].values())
            smoothed_tt = running_average(mean_tt, last_n=self.smooth_by)
            axes[0].plot(episodes, smoothed_tt, color=self.colors[idx], label=od, linewidth=self.line_width)
        for phase_idx, phase in enumerate(self.phases):
            color = self.phase_colors[phase_idx % len(self.phase_colors)]
            axes[0].axvline(x=phase,
                            label=self.phase_names[phase_idx],
                            linestyle='--',
                            color=color,
                            linewidth=self.line_width)
        axes[0].tick_params(axis='both', which='major', labelsize=self.tick_label_size)
        axes[0].set_xlabel('Episode', fontsize=self.label_size)
        axes[0].set_ylabel('Mean Travel Time', fontsize=self.label_size)
        axes[0].grid(True, axis='y')
        axes[0].set_title('Mean Human Travel Times Per OD', fontsize=self.title_size, fontweight='bold')
        axes[0].legend()

        # Plot variance travel times for all, humans and machines
        variance_travel_times = self._retrieve_data_per_kind(kc.TRAVEL_TIME, transform='variance')
        for idx, (kind, ep_tt_dict) in enumerate(variance_travel_times.items()):
            episodes = list(ep_tt_dict.keys())
            var_tts = list(ep_tt_dict.values())
            smoothed_var_tts = running_average(var_tts, last_n=self.smooth_by)
            axes[1].plot(episodes,
                         smoothed_var_tts,
                         color=self.colors[idx],
                         label=kind,
                         linewidth=self.line_width)

        for phase_idx, phase in enumerate(self.phases):
            color = self.phase_colors[phase_idx % len(self.phase_colors)]
            axes[1].axvline(x=phase,
                            label=self.phase_names[phase_idx],
                            linestyle='--',
                            color=color,
                            linewidth=self.line_width)
        axes[1].tick_params(axis='both', which='major', labelsize=self.tick_label_size)
        axes[1].set_xlabel('Episode', fontsize=self.label_size)
        axes[1].set_ylabel('Variance', fontsize=self.label_size)
        axes[1].grid(True, axis='y')
        axes[1].set_title('Variance Travel Times', fontsize=self.title_size, fontweight='bold')
        axes[1].legend()

        # Plot boxplot and violin plot for rewards
        all_travel_times = self._retrieve_data_per_kind(kc.TRAVEL_TIME)
        eps_to_plot = [ep-1 for ep in self.phases[1:]] + [self.saved_episodes[-1]]
        data_to_plot = [all_travel_times[kc.TYPE_HUMAN][ep] for ep in eps_to_plot]
        labels = [f'Humans\n({ph})' for ph in self.phase_names]

        bplot = axes[2].boxplot(data_to_plot, labels=labels, patch_artist=True)
        for idx, (patch, med) in enumerate(zip(bplot['boxes'], bplot['medians'])):
            color = self.colors[idx]
            patch.set_facecolor(color)
            med.set_color('black')
            med.set_linewidth(2)
        axes[2].tick_params(axis='both', which='major', labelsize=self.tick_label_size)
        axes[2].grid(axis = 'y')
        axes[2].set_ylabel('Travel Times', fontsize=self.label_size)
        axes[2].set_title(f'Human T.T. Distributions (End of Phases)', fontsize=self.title_size, fontweight='bold')

        dark_gray = '#333333'
        axes[3].set_facecolor(dark_gray)
        for idx, (label, data) in enumerate(zip(labels, data_to_plot)):
            data = np.array(data)
            data[np.isinf(data)] = np.nan  # Convert inf to NaN

            sns.kdeplot(data,
                        ax=axes[3],
                        label=label,
                        alpha=0.8,
                        fill=True,
                        linewidth=3,
                        color=self.colors[idx],
                        clip=(0, None))
            median_val, mean_val = np.nanmedian(data), np.nanmean(data)
            # Plot a vertical line from top to mid-plot for median
            axes[3].axvline(median_val,
                            color=self.colors[idx],
                            linestyle='-',
                            linewidth=2,
                            ymin=0.5,
                            ymax=1,
                            label=f'Median {label}')
            # Plot a vertical line from bottom to mid-plot for mean
            axes[3].axvline(mean_val,
                            color=self.colors[idx],
                            linestyle='--',
                            linewidth=2,
                            ymin=0,
                            ymax=0.5,
                            label=f'Mean {label}')
        axes[3].tick_params(axis='both', which='major', labelsize=self.tick_label_size)
        axes[3].set_xlim(0, None)
        axes[3].set_xlabel('Travel Times', fontsize=self.label_size)
        axes[3].set_ylabel('Probability Density', fontsize=self.label_size)
        axes[3].set_title(f'Human T.T. Distributions (End of Phases)', fontsize=self.title_size, fontweight='bold')
        axes[3].legend()

        plt.savefig(save_to)
        plt.close()
        logging.info(f"[SUCCESS] Travel time distributions are saved to {save_to}")

    ################################
    ########### ACTIONS ############
    ################################

    def visualize_actions(self) -> None:
        """Visualize the actions taken by the agent.

        Returns:
            None
        """

        save_to = make_dir(self.params[kc.PLOTS_FOLDER], kc.ACTIONS_PLOT_FILE_NAME)

        all_actions = self._retrieve_data_per_od(kc.ACTION)
        unique_actions = {
            od: {
                item for sublist in val.values() for item in sublist
            }
            for od, val in all_actions.items()
        }
        all_actions = {od : [Counter(a) for a in val.values()] for od, val in all_actions.items()}
        all_actions = {k: v for k,v in sorted(all_actions.items(), key=lambda x: f"{x[0]}")}
        num_od_pairs = len(all_actions)
        
        # Determine the layout of the subplots (rows x columns)
        num_columns = self.default_num_columns if num_od_pairs > self.default_num_columns else num_od_pairs
        num_rows = (num_od_pairs + num_columns - 1) // num_columns  # Calculate rows needed
        
        figure_size = (self.multimode_width * num_columns, self.multimode_height * num_rows)
        fig, axes = plt.subplots(num_rows, num_columns, figsize=figure_size)
        fig.tight_layout(pad=5.0)
        
        if num_rows > 1:
            axes = axes.flatten()   # Flatten axes
        if not hasattr(axes, '__getitem__'):
            axes = np.array([axes])  # If only one subplot

        for idx, (od, actions) in enumerate(all_actions.items()):
            ax = axes[idx]
            for idx2, unique_action in enumerate(unique_actions[od]):
                action_data = [ep_actions.get(unique_action, 0) for ep_actions in actions]
                action_data = running_average(action_data, last_n=self.smooth_by)
                ax.plot(self.saved_episodes,
                        action_data,
                        color=self.colors[idx2],
                        label=f"{unique_action}",
                        linewidth=self.line_width)

            for phase_idx, phase in enumerate(self.phases):
                color = self.phase_colors[phase_idx % len(self.phase_colors)]
                ax.axvline(x=phase,
                           label=self.phase_names[phase_idx],
                           linestyle='--',
                           color=color,
                           linewidth=self.line_width)

            ax.set_title(f"Actions for {od}", fontsize=self.title_size, fontweight='bold')
            ax.set_xlabel('Episodes', fontsize=self.label_size)
            ax.set_ylabel('Number of Drivers', fontsize=self.label_size)
            ax.grid(True, axis='y')
            ax.legend()
            ax.tick_params(axis='both', which='major', labelsize=self.tick_label_size)

        # Hide unused subplots if any
        for ax in axes[idx+1:]:
            ax.axis('off')

        for ax in axes.flat:
            ax.legend().set_visible(False)
            
        handles, labels = axes[0].get_legend_handles_labels()
        legend_loc = 'upper center' if num_od_pairs > 1 else 'upper right'
        fig.legend(handles, labels, loc=legend_loc, ncol=2, fontsize=self.legend_font_size)
        fig.subplots_adjust(top=0.90)

        plt.savefig(save_to)
        plt.close()
        logging.info(f"[SUCCESS] Actions are saved to {save_to}")

    ################################
    ######## ACTION SHIFTS #########
    ################################

    def visualize_action_shifts(self) -> None:
        """Visualize the action shifts taken by the agent.

        Returns:
            None
        """

        save_to = make_dir(self.params[kc.PLOTS_FOLDER], kc.ACTIONS_SHIFTS_PLOT_FILE_NAME)

        all_od_pairs = self._retrieve_all_od_pairs()
        all_od_pairs = sorted(all_od_pairs, key=lambda x: f"{x[0]}-{x[1]}")
        
        # Determine the layout of the subplots (rows x columns)
        num_columns = self.default_num_columns if len(all_od_pairs) > self.default_num_columns else len(all_od_pairs)
        num_rows = (len(all_od_pairs) + num_columns - 1) // num_columns  # Calculate rows needed
        
        figure_size = (self.multimode_width * num_columns, self.multimode_height * num_rows)
        fig, axes = plt.subplots(num_rows, num_columns, figsize=figure_size)
        fig.tight_layout(pad=5.0)
        
        if num_rows > 1:   axes = axes.flatten()   # Flatten axes
        if not hasattr(axes, '__getitem__'):    axes = np.array([axes])  # If only one subplot

        for idx, od in enumerate(all_od_pairs):
            ax = axes[idx]
            origin, destination = od

            all_actions, unique_actions = self._retrieve_selected_actions(origin, destination)

            for idx2, (kind, ep_to_actions) in enumerate(all_actions.items()):

                episodes = list(ep_to_actions.keys())

                for idx3, action in enumerate(unique_actions):
                    action_data = [ep_ctr[action] / sum(ep_ctr.values()) for ep_ctr in all_actions[kind].values()]
                    action_data = running_average(action_data, last_n=self.smooth_by)
                    color = self.colors[idx3]
                    linestyle = self.linestyles[idx2 % len(self.linestyles)]
                    ax.plot(episodes,
                            action_data,
                            color=color,
                            linestyle=linestyle,
                            label=f"{kind}-{action}",
                            linewidth=self.line_width)

            for phase_idx, phase in enumerate(self.phases):
                color = self.phase_colors[phase_idx % len(self.phase_colors)]
                ax.axvline(x=phase,
                           label=self.phase_names[phase_idx],
                           linestyle='--',
                           color=color,
                           linewidth=self.line_width)

            ax.set_xlabel('Episodes', fontsize=self.label_size)
            ax.set_ylabel('Fraction of drivers', fontsize=self.label_size)
            ax.grid(True, axis='y')
            ax.set_title(f'Actions for {od}', fontsize=self.title_size, fontweight='bold')
            ax.legend()
            ax.tick_params(axis='both', which='major', labelsize=self.tick_label_size)

        for ax in axes[idx+1:]:
            ax.axis('off')    # Hide unused subplots if any
        for ax in axes.flat:
            ax.legend().set_visible(False)
        handles, labels = axes[0].get_legend_handles_labels()
        legend_loc = 'upper center' if len(all_od_pairs) > 1 else 'upper right'
        fig.legend(handles, labels, loc=legend_loc, ncol=3, fontsize=self.legend_font_size)
        fig.subplots_adjust(top=0.85)

        plt.savefig(save_to)
        plt.close()
        logging.info(f"[SUCCESS] Actions shifts are saved to {save_to}")

    ################################
    ########## SIM LENGTH ##########
    ################################

    def visualize_sim_length(self) -> None:
        """Visualize the simulation length.

        Returns:
            None
        """

        save_to = make_dir(self.params[kc.PLOTS_FOLDER], kc.SIM_LENGTH_PLOT_FILE_NAME)
        sim_lengths = self._retrieve_sim_length()
        sim_lengths = running_average(sim_lengths, last_n=self.smooth_by)

        plt.figure(figsize=(self.default_width, self.default_height), layout='tight')
        plt.plot(self.saved_episodes,
                 sim_lengths,
                 color=self.colors[0],
                 label="Simulation time steps",
                 linewidth=self.line_width)
        for phase_idx, phase in enumerate(self.phases):
            color = self.phase_colors[phase_idx % len(self.phase_colors)]
            plt.axvline(x=phase,
                        label=self.phase_names[phase_idx],
                        linestyle='--',
                        color=color,
                        linewidth=self.line_width)
        plt.xticks(fontsize=self.tick_label_size)
        plt.yticks(fontsize=self.tick_label_size)
        plt.xlabel('Episode', fontsize=self.label_size)
        plt.ylabel('Simulation Length', fontsize=self.label_size)
        plt.title('Simulation Length Over Episodes', fontsize=self.title_size, fontweight='bold')
        plt.legend(fontsize=self.legend_font_size)

        plt.savefig(save_to)
        plt.close()
        logging.info(f"[SUCCESS] Simulation lengths are saved to {save_to}")

    def _retrieve_sim_length(self) -> list:
        """Retrieve the simulation length.

        Returns:
            latest arrivals (list): List of the latest arrivals in the simulation.
        """

        latest_arrivals = list()
        for episode in self.saved_episodes:
            data_path = os.path.join(self.episodes_folder, f"ep{episode}.csv")
            data = pd.read_csv(data_path)
            arrival_times = data.apply(lambda row: row[kc.AGENT_START_TIME] + row[kc.TRAVEL_TIME], axis=1)
            latest_arrival = max(arrival_times)
            latest_arrivals.append(latest_arrival)
        return latest_arrivals

    ################################
    ############ LOSES #############
    ################################

    def visualize_losses(self) -> None:
        """Visualize the losses.

        Returns:
            None
        """

        save_to = make_dir(self.params[kc.PLOTS_FOLDER], kc.LOSSES_PLOT_FILE_NAME)

        losses = self._retrieve_losses()
        if not losses: return
        losses = running_average(losses, last_n=self.smooth_by)

        plt.figure(figsize=(self.default_width, self.default_height), layout='tight')
        plt.plot(losses, color=self.colors[0], linewidth=self.line_width)
        plt.xlabel('Training Progress (Episodes)', fontsize=self.label_size)
        plt.ylabel('MSE Loss (Log Scale)', fontsize=self.label_size)
        plt.title('Mean MSE Loss Over Training Progress for AVs', fontsize=self.title_size, fontweight='bold')
        plt.yscale('log')
        plt.grid(True, axis='y')
        plt.xticks(fontsize=self.tick_label_size)
        plt.yticks(fontsize=self.tick_label_size)

        plt.savefig(save_to)
        plt.close()
        logging.info(f"[SUCCESS] Losses are saved to {save_to}")

    def _retrieve_losses(self) -> Union[list[float], None]:
        """Retrieve the losses.

        Returns:
            losses list[float]: List of the losses.
        """

        losses = list()
        if not os.path.isfile(self.loss_file_path):
            return None
        with open(self.loss_file_path, "r") as file:
            lines = file.readlines()
            for line in lines:
                losses.append(float(line.strip()))
        return losses

    ################################
    ########### HELPERS ############
    ################################

    def _retrieve_data_per_kind(self, data_key, transform=None) -> dict[str, dict]:
        """Retrieve data per kind.

        Args:
            data_key (str): The key of the data you want to retrieve.
            transform (callable, optional): Optional transform to be applied
        Returns:
            all_values_dict (dict[Any, dict]): Dictionary of all data values per kind.
        """

        all_values_dict = {kc.ALL : dict()}
        for episode in self.saved_episodes:
            data_path = os.path.join(self.episodes_folder, f"ep{episode}.csv")
            data = pd.read_csv(data_path)
            kinds, values, found_kinds = data[kc.AGENT_KIND], data[data_key], set(data[kc.AGENT_KIND])
            for kind in found_kinds:
                if kind not in all_values_dict:    all_values_dict[kind] = dict()
            values_per_kind =  {k : list() for k in found_kinds}
            for kind, value in zip(kinds, values):
                values_per_kind[kind].append(value)
            values_per_kind[kc.ALL] = [item for sublist in values_per_kind.values() for item in sublist]
            for kind in values_per_kind:
                if transform == 'mean':
                    all_values_dict[kind][episode] = mean(values_per_kind[kind])
                elif transform == 'variance':
                    all_values_dict[kind][episode] = variance(values_per_kind[kind])
                else:
                    all_values_dict[kind][episode] = values_per_kind[kind]
        return all_values_dict
    
    def _retrieve_data_per_od(self, data_key, transform=None) -> dict[str, dict]:
        """Retrieve data per od.

        Args:
            data_key (str): The key of the data you want to retrieve.
            transform (callable, optional): Optional transform to be applied
        Returns:
            all_values_dict (dict[str, dict]): Dictionary of all data values per kind.
        """

        all_od_pairs = self._retrieve_all_od_pairs()
        od_to_key = lambda o, d: f"({o}, {d})"
        all_values_dict = {od_to_key(od[0], od[1]) : dict() for od in all_od_pairs}
        for episode in self.saved_episodes:
            data_path = os.path.join(self.episodes_folder, f"ep{episode}.csv")
            episode_data = pd.read_csv(data_path)
            episode_origins, episode_destinations = episode_data[kc.AGENT_ORIGIN], episode_data[kc.AGENT_DESTINATION]
            values = episode_data[data_key]
            values_per_od = {od_to_key(od[0], od[1]): list() for od in all_od_pairs}
            for idx, value in enumerate(values):
                values_per_od[od_to_key(episode_origins[idx], episode_destinations[idx])].append(value)
            for od in values_per_od:
                if transform == 'mean':
                    all_values_dict[od][episode] = mean(values_per_od[od])
                elif transform == 'variance':
                    all_values_dict[od][episode] = variance(values_per_od[od])
                else:
                    all_values_dict[od][episode] = values_per_od[od]
        return all_values_dict

    def _retrieve_selected_actions(self, origin, destination) -> [dict, dict]:
        """Retrieve selected actions.

        Args:
            origin (str): The origin of the action.
            destination (str): The destination of the action.

        Returns:
            all_actions (dict): Dictionary of all selected actions.
            unique_actions (dict): Dictionary of all unique actions.
        """

        all_actions = dict()
        unique_actions = set()
        for episode in self.saved_episodes:
            data_path = os.path.join(self.episodes_folder, f"ep{episode}.csv")
            data = pd.read_csv(data_path)
            data = data[(data[kc.AGENT_ORIGIN] == origin) & (data[kc.AGENT_DESTINATION] == destination)]
            kinds, actions = data[kc.AGENT_KIND], data[kc.ACTION]
            for kind in set(kinds):
                if kind not in all_actions:    all_actions[kind] = dict()
            unique_actions.update(actions)
            actions_counters = {k : Counter() for k in set(kinds)}
            for kind, action in zip(kinds, actions):
                actions_counters[kind][action] += 1
            for kind in actions_counters.keys():
                all_actions[kind][episode] = actions_counters[kind]
        return all_actions, unique_actions

    def _retrieve_all_od_pairs(self):
        """Retrieve all OD pairs.

        Returns:
            all_od_pairs (dict): Dictionary of all OD pairs.
        """

        all_od_pairs = list()
        data_path = os.path.join(self.episodes_folder, f"ep{self.saved_episodes[0]}.csv")
        episode_data = pd.read_csv(data_path)
        episode_data = episode_data[[kc.AGENT_ORIGIN, kc.AGENT_DESTINATION]]
        for _, row in episode_data.iterrows():
            origin, destination = int(row[kc.AGENT_ORIGIN]), int(row[kc.AGENT_DESTINATION])
            all_od_pairs.append((origin, destination))
        all_od_pairs = list(set(all_od_pairs))
        return all_od_pairs


def plotter(params = None):
    """
    Creates a ``Plotter`` instance and plots the results of the training.

    Returns:
        plotter: plotter
    """
    if params is None:
        logging.warning(f"No parameters provided for plotter, "
                        f"using default parameters. This may result in incorrect plots.")
        curr_dir = os.path.dirname(os.path.realpath(__file__))
        params_path = os.path.join(curr_dir, f'../environment/{kc.PARAMS_FILE}')
        params = get_params(params_path)
        params = params[kc.PLOTTER]

    plotter = Plotter(params)
    plotter.plot()
    return plotter


import json
import os
import requests
import sys
import time

from prettytable import PrettyTable

from routerl.keychain import Keychain as kc


def get_params(file_path, resolve=True, update=None):
    # Read params.json, resolve dependencies
    params = read_json(file_path)
    if (update is not None) and isinstance(update, dict):
        update_params(params, update)
    params = resolve_ods(params)
    if resolve:
        params = resolve_param_dependencies(params)
    return params


def update_params(old_params: dict, new_params: dict):
    for key, value in new_params.items():
        if not (key in old_params):
            raise ValueError(f"Invalid parameter: {key}")
        elif isinstance(old_params.get(key), str) and old_params.get(key).startswith("${"):
            raise ValueError(f"Invalid parameter: {key}")
        elif (not isinstance(value, dict)) and isinstance(old_params.get(key), dict):
            raise ValueError(f"Can't update group parameter {key} with value: {value}")
        elif isinstance(value, dict) and isinstance(old_params.get(key), dict):
            update_params(old_params[key], value)
        else:
            old_params[key] = value


def resolve_ods(params):
    curr_dir = os.path.dirname(os.path.abspath(__file__))
    default_od_path = os.path.join(curr_dir, kc.DEFAULT_ODS_PATH)
    default_ods = read_json(default_od_path)[params[kc.SIMULATOR][kc.NETWORK_NAME]]
    if params[kc.PATH_GEN][kc.ORIGINS] == "default":
        params[kc.PATH_GEN][kc.ORIGINS] = default_ods[kc.ORIGINS]
    if params[kc.PATH_GEN][kc.DESTINATIONS] == "default":
        params[kc.PATH_GEN][kc.DESTINATIONS] = default_ods[kc.DESTINATIONS]
    if params[kc.SIMULATOR][kc.NETWORK_NAME] == kc.TWO_ROUTE_YIELD:
        params[kc.PATH_GEN][kc.NUMBER_OF_PATHS] = 2
    if params[kc.SIMULATOR][kc.NETWORK_NAME] == kc.MANHATTAN: # manhattan network is big so we store the network files in zenodo
        zenodo_record_id = kc.ZENODO_RECORD_ID
        api_url = f"https://zenodo.org/api/records/{zenodo_record_id}"

        curr_dir = os.path.dirname(os.path.abspath(__file__))
        save_folder = os.path.join(curr_dir, kc.NETWORK_FOLDER).replace("$net$", kc.MANHATTAN)
        
        response = requests.get(api_url)
        data = response.json()

        # Download all files
        for file_info in data["files"]:
            file_url = file_info["links"]["self"]
            file_name = file_info["key"]

            # Construct the full save path
            save_path = os.path.join(save_folder, file_name)
            response = requests.get(file_url, stream=True)
            
            # Write the file to the specified folder
            with open(save_path, "wb") as file:
                for chunk in response.iter_content(chunk_size=1024):
                    file.write(chunk)
    return params


def resolve_param_dependencies(params):  # Resolving dependent parameters in params.json
    for category, settings in params.items():
        for key, value in settings.items():
            if isinstance(value, str) and value.startswith("${") and value.endswith("}"):
                path = value[2:-1].split('.')  # Extract the reference path
                ref_value = params
                for step in path:
                    ref_value = ref_value.get(step, {})
                if not isinstance(ref_value, dict):  # Ensure it's not a nested structure
                    params[category][key] = ref_value
    return params


def confirm_env_variable(env_var, append=None):
    if env_var in os.environ:
        print("[CONFIRMED] Environment variable exists: %s" % env_var)
        if append:
            path = os.path.join(os.environ[env_var], append)
            sys.path.append(path)
            print("[SUCCESS] Added module directory: %s" % path)
    else:
        raise ImportError("Please declare the environment variable '%s'" % env_var)


def read_json(file_path):  # Read json file, return as dict
    try:
        with open(file_path, 'r') as file:
            file_data = json.load(file)
    except FileNotFoundError:
        print(f"[ERROR] Cannot locate: %s" % (file_path))
        raise
    return file_data


def make_dir(folders, filename=None):  # Make dir if not exists, make full path
    if not folders:
        print("[ERROR] Expected at least one folder name.")
        raise ValueError
    if not isinstance(folders, list):
        folders = [folders]
    path = str()
    for folder_name in folders:
        path = os.path.join(path, folder_name)
        if not os.path.exists(path):
            os.mkdir(path)
    if filename: path = os.path.join(path, filename)
    return path


def show_progress_bar(message, start_time, progress, target, end_line=''):  # Just printing progress bar with ETA
    bar_length = 50
    progress_fraction = progress / target
    filled_length = int(bar_length * progress_fraction)
    bar = 'X' * filled_length + '-' * (bar_length - filled_length)
    elapsed_time = time.time() - start_time
    remaining_time = ((elapsed_time / progress_fraction) - elapsed_time) if progress_fraction else 0
    remaining_time = time.strftime("%H:%M:%S", time.gmtime(remaining_time))
    print(f'\r[%s]: |%s| %.2f%%, ETA: %s' % (message.upper(),
                                             bar,
                                             progress_fraction * 100,
                                             remaining_time),
          end=end_line)


def show_progress(name_of_operation, start_time, progress, target, end_line=''):  # Just printing progress bar with ETA
    progress_fraction = progress / target
    elapsed_time = time.time() - start_time
    remaining_time = ((elapsed_time / progress_fraction) - elapsed_time) if progress_fraction else 0
    remaining_time = time.strftime("%H:%M:%S", time.gmtime(remaining_time))
    print(f'\r[%s PROGRESS]: %.2f%%, ETA: %s' % (name_of_operation.upper(),
                                                 progress_fraction * 100,
                                                 remaining_time),
          end=end_line)


def remove_double_quotes(text):
    text = str(text).replace('"', '')
    return text


def list_to_string(from_list, separator=', '):
    out_str = ""
    first_time = True
    for item in from_list:
        if first_time:
            out_str = str(item)
            first_time = False
        else:
            out_str = "%s%s%s" % (out_str, separator, item)

    return out_str


def string_to_list(text, seperator, brackets=False):
    if brackets:
        text = text.strip("[]")
    elements = text.split(seperator)
    return elements


def df_to_prettytable(df, header_message=None, print_every=1):
    table = PrettyTable()
    table.field_names = df.columns.tolist()
    for index, row in df.iterrows():
        if not (index % print_every):
            table.add_row(row.tolist())
    if header_message:
        print(f"##### {header_message} #####")
    print(table)


def running_average(values, last_n=0):
    # last_n -> -1 disables the averaging, 0 averages all, n averages the last n
    if last_n < 0: return values

    running_sum, running_averages = 0, list()
    for idx, _ in enumerate(values):
        start_from = max(0, idx - last_n) if last_n > 0 else 0
        divide_by = idx - start_from + 1
        running_sum = sum(values[start_from:idx + 1])
        running_averages.append(running_sum / divide_by)
    return running_averages
